"""grouch.context

Provides the TypecheckContext class, instances of which are carried around
throughout a type-checking traversal of an entire obejct graph.
"""

# created 2001/02, GPW
#
# Copyright (c) 2001 CNRI.  All Rights Reserved.

__revision__ = "$Id: context.py,v 1.13 2001/09/05 14:21:23 gward Exp $"

import sys, string
from copy import copy


class TypecheckContext:
    """
    Stores preferences, context, and results of a type-checking traversal
    over an object graph.

    Instance attributes:
      context : [(action:string, label:string)]
        list of tuples that record each step through the object graph.  The
        tuples look like eg. ('item', 3) for looking up item 3 of a
        sequence, or ('attr', 'foo') for looking up attribute 'foo' of an
        instance.  The first tuple in the list might look like ('name',
        'start') to distinguish it from an attribute lookup, which it
        probably isn't.  The context list is used to generate meaningful
        error messages from deep within an object graph.
      object_seen : { int : boolean }
        map object IDs to true values for objects that have already
        been visited and checked
      mode : string   # one of 'memory', 'zodb'
        what type of object graph we're traversing; currently used to
        know how to determine object IDs
      errors : [ (context, message) ]
        list of typechecking errors.  'context' is a context list (just a
        copy of the 'context' attribute at the point where the error is
        seen).  'message' is a human-readable string describing the type
        error.
      report_errors : boolean = true
        if true, errors will be reported (printed to stderr) as soon
        as they are seen.  (Errors are always stored in the 'errors'
        list for future reference.)
      visit_counts : { metatype:string : { extra:string : count:int } }
        counts of the types of objects visited in a type-checking
        traversal.  The keys of 'visit_counts' are strings like
        "atomic" or "instance" -- the major "meta-types" of the
        Grouch type system.  The values are dictionaries mapping
        specific type names (eg. "int" or "string" are specific atomic
        types; "Foo" and "Bar" might be specific instance types)
        to the number of objects of that type visited.  Certain
        metatypes (eg. boolean, any) won't have any extra type info,
        so their sub-dictionaries will be singleton whose sole key
        is None.
    """

    def __init__ (self, mode='memory', start_name=None, report_errors=1):
        if mode not in ('memory', 'zodb'):
            raise ValueError, "unknown mode: %s" % mode

        self.context = []
        self.start_context(start_name)

        self.object_seen = {}
        self.mode = mode
        self.errors = []
        self.report_errors = report_errors
        self.visit_counts = {}

        if self.mode == 'zodb':
            self._id = self._zodb_id
        else:
            self._id = id

    def _zodb_id (self, object):
        try:
            return object._p_oid
        except AttributeError:
            return id(object)           # not great, but better than nothing

    def start_context (self, start_name):
        del self.context[:]
        if start_name:
            self.context.append(('name', start_name))

    def reset (self, start_name=None):
        del self.context[:]
        self.object_seen.clear()
        del self.errors[:]
        self.start_context(start_name)
        self.visit_counts.clear()


    def has_seen (self, object):
        return self.object_seen.get(self._id(object))

    def set_seen (self, object, seen=1):
        self.object_seen[self._id(object)] = seen

    def check_seen (self, object):
        """If 'object' has already been visited, return true; otherwise,
        record that the object has been visited and return false.
        """
        oid = self._id(object)
        if self.object_seen.get(oid):
            return 1
        else:
            self.object_seen[oid] = 1
            return 0

    def push_context (self, action, label):
        """Descend a level in the object graph and append a new
        entry to the 'context' list.
        """
        if action not in ('attr', 'item', 'dkey', 'dval'):
            raise ValueError, "invalid 'action': must be 'attr' or 'item'"
        self.context.append((action, label))

    def format_context (self):
        context_l = []
        for (action, label) in self.context:
            if action == 'name':
                context_l.append(label)
            elif action == 'attr':
                if context_l:
                    context_l.append('.' + label)
                else:
                    context_l.append(label)
            elif action in ('item', 'dkey', 'dval'):
                context_l.append('[%s]' % `label`)
                if action == 'dkey':
                    context_l.append(' (key)')

        return string.join(context_l, '')

    def replace_context (self, action, label):
        del self.context[-1]
        self.push_context(action, label)

    def pop_context (self):
        if len(self.context) < 1:
            raise RuntimeError, "can't pop context: already at top"
        del self.context[-1]


    def add_error (self, message):
        context = self.format_context()
        self.errors.append((context, message))
        if self.report_errors:
            self.write_error(sys.stderr, self.errors[-1])

    def forget_errors (self, num=1):
        if num > 0:
            del self.errors[-num:len(self.errors)]
            
    def format_errors (self):
        errors = []
        for (context, message) in self.errors:
            if context:
                errors.append("%s: %s" % (context, message))
            else:
                errors.append(message)
        return errors

    def write_error (self, file, error):
        (context, message) = error
        if context:
            file.write(context + ":\n")
            file.write("  " + message + "\n")
        else:
            file.write(message + "\n")

    def write_errors (self, file):
        for error in self.errors:
            self.write_error(file, error)

    def num_errors (self):
        return len(self.errors)


    def count_object (self, type, extra_type=None):
        typename = type.metatype
        counts = self.visit_counts.get(typename)
        if counts is None:
            counts = self.visit_counts[typename] = {}
        if counts.has_key(extra_type):
            counts[extra_type] += 1
        else:
            counts[extra_type] = 1

    def uncount_object (self, type, extra_type=None):
        typename = type.metatype
        if self.visit_counts[typename][extra_type] == 1:
            del self.visit_counts[typename][extra_type]
            if len(self.visit_counts[typename]) == 0:
                del self.visit_counts[typename]
        else:
            self.visit_counts[typename][extra_type] -= 1

    def report_counts (self, file=None):
        if file is None:
            file = sys.stdout
        metatypes = self.visit_counts.keys()
        metatypes.sort()

        for metatype in metatypes:
            counts = self.visit_counts[metatype]
            if len(counts) == 1 and counts.keys()[0] is None:
                file.write("%-10s%-55s%10d\n" %
                           (metatype, "", counts[None]))
            else:
                file.write("%s:\n" % metatype)
                types = counts.keys()
                types.sort()
                for type in types:
                    file.write("%2s%-63s%10d\n" %
                               ("", type, counts[type]))

        print "objects visited:", len(self.object_seen)

# class TypecheckContext



"""grouch.type_parser

SPARK-based scanner and parser for Grouch's type language.
"""

# created 2001/03/15, GPW
#
# Copyright (c) 2001 CNRI.  All Rights Reserved.

__revision__ = "$Id: type_parser.py,v 1.7 2001/09/05 14:21:23 gward Exp $"


import string
from types import StringType

from grouch.spark import GenericScanner, GenericParser
from grouch.schema import ObjectSchema
from grouch.util import get_class_object
from grouch import valuetype


# 
# Grammar for types:
# type : NAME                     # atomic, alias, or instance type
#      | container_type           # list, tuple, dictionary
#      | NAME container_type      # instance-container type
#      | union_type
#
# container_type : list_type
#                | tuple_type
#                | dictionary_type
# list_type      : "[" type "]"
# tuple_type     : "(" (type ",")* type "*"? ","? ")"
# dictionary_type: "{" type ":" type "}"
#
# union_type : type ("|" type)+
# 
# Terminals (tokens):
#   NAME : [a-zA-Z_][a-zA-Z0-9_]*(\.[a-zA-Z_][a-zA-Z0-9_]*)*
#   "[" "]"
#   "(" ")" ","
#   "{" "}" ":"
#   "|"


class Name:
    def __init__ (self, name):
        self.name = name

    def __cmp__ (self, other):
        if other == "NAME":             # for the parser below
            return 0
        elif type(other) is StringType:
            return 1
        elif isinstance(other, Name):   # for comparing names
            return cmp(self.name, other.name)
        else:
            raise TypeError, "can't compare Name instance to %s" % `other`

    def __repr__ (self):
        return "<Name at %x: %s>" % (id(self), self.name)

    def __str__ (self):
        return self.name


class TypeScanner (GenericScanner):

    def tokenize (self, input):
        self.tokens = []
        GenericScanner.tokenize(self, input)
        return self.tokens

    def t_whitespace (self, s):
        r'\s+'
        pass

    def t_name (self, s):
        r'[a-zA-Z_][a-zA-Z0-9_]*(\.[a-zA-Z_][a-zA-Z0-9_]*)*'
        self.tokens.append(Name(s))

    def t_punct (self, s):
        r'[\[\]\(\)\{\}\,\:\|\*]'
        self.tokens.append(s)


    # Looks like errors can be handled two ways:
    #   * have a 't_default()' that recognizes nothing, and an
    #     'error()' that raises an exception with 'pos'
    #   * have a 't_default()' that accepts anything and then
    #     raises an exception
    # The first gives you the position where the error occurs,
    # the second gives you the text that causes the error.

    #def t_default (self, s):
    #    r'.'
    #    raise ValueError, "invalid token %s" % `s`

    def t_default (self, s):
        pass
    
    def error (self, s, pos):
        raise ValueError, "invalid token at position %d: '%s'" % (pos, s[pos:])


class TypeParser (GenericParser):

    def __init__ (self, start="type", schema=None):
        GenericParser.__init__(self, start)
        self.schema = schema

    def set_schema (self, schema):
        self.schema = schema

    def parse (self, tokens):
        if self.schema is None:
            raise RuntimeError, "schema not set"
        return GenericParser.parse(self, tokens)

    def error (self, token):
        raise ValueError, "Syntax error at or near '%s'" % token

    def p_type_1 (self, (name,)):
        'type ::= NAME'
        #print "type_1: got a name: %s" % `name`
        return self.schema.get_type(str(name))
    
    def p_type_2 (self, args):
        'type ::= container_type'
        #print "type_2: args =", args
        assert isinstance(args[0], valuetype.ContainerType)
        return args[0]

    def p_type_3 (self, args):
        'type ::= NAME container_type'
        (name, ctype) = args
        #print "type_3: got an instance-container type:", args

        name = str(name)
        # Check if this is really an alias for something (presumably a class!)
        alias_val = self.schema.get_alias(name)
        if alias_val:
            if not alias_val.is_plain_instance_type():
                raise ValueError, \
                      ("%s is an alias, but not for a plain instance type"
                       % name)
            klass_name = alias_val.klass_name
        else:
            # this should become ObjectSchema.make_instance_container_type()
            klass_name = name

        ictype = valuetype.InstanceContainerType(self.schema)
        ictype.set_class_name(klass_name)
        ictype.set_container_type(ctype)
        return ictype

    def p_type_4 (self, args):
        'type ::= union_type'
        typelist = args[0]
        #print "type_4: got a union type:", typelist
        utype = valuetype.UnionType(self.schema)
        utype.set_union_types(typelist)
        return utype


    # Container types are about the only interesting syntax in the whole
    # grammar.  Of those, list and dictionary types are pretty simple:
    #   list_type      : "[" type "]"
    #   dictionary_type: "{" type ":" type "}"
    # So they are implemented directly as alternate productions
    # for the 'container_type' terminal.
    #
    # Tuple types are a tad more interesting.  In extended BNF:
    #   tuple_type     : "(" (type ",")* type "*"? ","? ")"
    # which refactors into vanilla BNF as:
    #   tuple_type : "(" type_list tl_tail ")"
    #   type_list  : type "," type_list
    #              | type
    #   tl_tail    : ","
    #              | "*"
    #              | "," "*"
    #              | null

    def p_container_type_1 (self, args): # list type
        'container_type ::= [ type ]'
        (_, eltype, _) = args
        #print "container_type_1: got a list type: %s" % `eltype`
        return self.schema.make_list_type(eltype)

    def p_container_type_2 (self, args): # tuple type
        'container_type ::= ( type_list tl_tail )'
        (_, eltypes, extended, _) = args
        #print "container_type_3: got a tuple type: %s (extended=%s)" % \
        #      (map(repr, eltypes), extended)
        return self.schema.make_tuple_type(eltypes, extended)

    def p_container_type_3 (self, args): # dictionary type
        'container_type ::= { type : type }'
        (_, keytype, _, valtype, _) = args
        #print "container_type_3: got a dict type: %s -> %s" % \
        #      (`keytype`, `valtype`)
        return self.schema.make_dictionary_type(keytype, valtype)


    # Grunt-work productions for parsing tuples

    def p_type_list_1 (self, args):
        'type_list ::= type , type_list'
        (type, _, typelist) = args
        typelist.insert(0, type)
        return typelist

    def p_type_list_2 (self, args):
        'type_list ::= type'
        return [args[0]]

    def p_tl_tail_1 (self, args):
        'tl_tail ::= ,'
        return 0                        # not an extended tuple
    
    def p_tl_tail_2 (self, args):
        'tl_tail ::= *'
        return 1                        # extended tuple
    
    def p_tl_tail_3 (self, args):
        'tl_tail ::= * ,'
        return 1                        # ditto

    def p_tl_tail_4 (self, args):
        'tl_tail ::='
        return 0                        # not extended


    # Finally, union types.  In EBNF:
    #   union_type : type ("|" type)+
    # which refactors to
    #   union_type     : type "|" union_type
    #   union_type     : type "|" type
    
    def p_union_type_1 (self, args):
        'union_type ::= type | union_type'
        (type, _, typelist) = args
        typelist.insert(0, type)
        return typelist
    
    def p_union_type_2 (self, args):
        'union_type ::= type | type'
        (type1, _, type2) = args
        return [type1, type2]


def get_class_def (schema, name):
    klass_def = schema.get_class_definition(name)
    if not klass_def:
        raise ValueError, \
              "invalid class %s: no class definition in schema" % `name`
    return klass_def



class Thing:
    """A single thing, which may be an animal, vegetable, or mineral.
    The only property common to all things is a name.

    Instance attributes:
      name : string
        the name of the thing
    """

    def __init__ (self, name):
        self.name = name

    def set_name (self, name):
        return self.name


class ThingCollection:
    """A collection of many things, indexed by name.

    Instance attributes:
      things : { string : Thing }
        all the things in the collection

    """

    def __init__ (self):
        self.things = {}

    def add_thing (self, thing):
        self.things[thing.name] = thing



#  Copyright (c) 1998-2000 John Aycock
#  
#  Permission is hereby granted, free of charge, to any person obtaining
#  a copy of this software and associated documentation files (the
#  "Software"), to deal in the Software without restriction, including
#  without limitation the rights to use, copy, modify, merge, publish,
#  distribute, sublicense, and/or sell copies of the Software, and to
#  permit persons to whom the Software is furnished to do so, subject to
#  the following conditions:
#  
#  The above copyright notice and this permission notice shall be
#  included in all copies or substantial portions of the Software.
#  
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

__version__ = 'SPARK-0.6.1'

import re
import sys
import string

def _namelist(instance):
	namelist, namedict, classlist = [], {}, [instance.__class__]
	for c in classlist:
		for b in c.__bases__:
			classlist.append(b)
		for name in dir(c):
			if not namedict.has_key(name):
				namelist.append(name)
				namedict[name] = 1
	return namelist

class GenericScanner:
	def __init__(self):
		pattern = self.reflect()
		self.re = re.compile(pattern, re.VERBOSE)

		self.index2func = {}
		for name, number in self.re.groupindex.items():
			self.index2func[number-1] = getattr(self, 't_' + name)

	def makeRE(self, name):
		doc = getattr(self, name).__doc__
		rv = '(?P<%s>%s)' % (name[2:], doc)
		return rv

	def reflect(self):
		rv = []
		for name in _namelist(self):
			if name[:2] == 't_' and name != 't_default':
				rv.append(self.makeRE(name))

		rv.append(self.makeRE('t_default'))
		return string.join(rv, '|')

	def error(self, s, pos):
		print "Lexical error at position %s" % pos
		raise SystemExit

	def tokenize(self, s):
		pos = 0
		n = len(s)
		while pos < n:
			m = self.re.match(s, pos)
			if m is None:
				self.error(s, pos)

			groups = m.groups()
			for i in range(len(groups)):
				if groups[i] and self.index2func.has_key(i):
					self.index2func[i](groups[i])
			pos = m.end()

	def t_default(self, s):
		r'( . | \n )+'
		pass

class GenericParser:
	def __init__(self, start):
		self.rules = {}
		self.rule2func = {}
		self.rule2name = {}
		self.collectRules()
		self.startRule = self.augment(start)
		self.ruleschanged = 1

	_START = 'START'
	_EOF = 'EOF'

	#
	#  A hook for GenericASTBuilder and GenericASTMatcher.
	#
	def preprocess(self, rule, func):	return rule, func

	def addRule(self, doc, func):
		rules = string.split(doc)

		index = []
		for i in range(len(rules)):
			if rules[i] == '::=':
				index.append(i-1)
		index.append(len(rules))

		for i in range(len(index)-1):
			lhs = rules[index[i]]
			rhs = rules[index[i]+2:index[i+1]]
			rule = (lhs, tuple(rhs))

			rule, fn = self.preprocess(rule, func)

			if self.rules.has_key(lhs):
				self.rules[lhs].append(rule)
			else:
				self.rules[lhs] = [ rule ]
			self.rule2func[rule] = fn
			self.rule2name[rule] = func.__name__[2:]
		self.ruleschanged = 1

	def collectRules(self):
		for name in _namelist(self):
			if name[:2] == 'p_':
				func = getattr(self, name)
				doc = func.__doc__
				self.addRule(doc, func)

	def augment(self, start):
		#
		#  Tempting though it is, this isn't made into a call
		#  to self.addRule() because the start rule shouldn't
		#  be subject to preprocessing.
		#
		startRule = (self._START, ( start, self._EOF ))
		self.rule2func[startRule] = lambda args: args[0]
		self.rules[self._START] = [ startRule ]
		self.rule2name[startRule] = ''
		return startRule

	def makeFIRST(self):
		union = {}
		self.first = {}
		
		for rulelist in self.rules.values():
			for lhs, rhs in rulelist:
				if not self.first.has_key(lhs):
					self.first[lhs] = {}

				if len(rhs) == 0:
					self.first[lhs][None] = 1
					continue

				sym = rhs[0]
				if not self.rules.has_key(sym):
					self.first[lhs][sym] = 1
				else:
					union[(sym, lhs)] = 1
		changes = 1
		while changes:
			changes = 0
			for src, dest in union.keys():
				destlen = len(self.first[dest])
				self.first[dest].update(self.first[src])
				if len(self.first[dest]) != destlen:
					changes = 1

	#
	#  An Earley parser, as per J. Earley, "An Efficient Context-Free
	#  Parsing Algorithm", CACM 13(2), pp. 94-102.  Also J. C. Earley,
	#  "An Efficient Context-Free Parsing Algorithm", Ph.D. thesis,
	#  Carnegie-Mellon University, August 1968, p. 27.
	#
	
	def typestring(self, token):
		return None

	def error(self, token):
		print "Syntax error at or near `%s' token" % token
		raise SystemExit

	def parse(self, tokens):
		tree = {}
		tokens.append(self._EOF)
		states = { 0: [ (self.startRule, 0, 0) ] }
		
		if self.ruleschanged:
			self.makeFIRST()

		for i in xrange(len(tokens)):
			states[i+1] = []

			if states[i] == []:
				break				
			self.buildState(tokens[i], states, i, tree)

		#_dump(tokens, states)

		if i < len(tokens)-1 or states[i+1] != [(self.startRule, 2, 0)]:
			del tokens[-1]
			self.error(tokens[i-1])
		rv = self.buildTree(tokens, tree, ((self.startRule, 2, 0), i+1))
		del tokens[-1]
		return rv

	def buildState(self, token, states, i, tree):
		needsCompletion = {}
		state = states[i]
		predicted = {}
		
		for item in state:
			rule, pos, parent = item
			lhs, rhs = rule

			#
			#  A -> a . (completer)
			#
			if pos == len(rhs):
				if len(rhs) == 0:
					needsCompletion[lhs] = (item, i)

				for pitem in states[parent]:
					if pitem is item:
						break

					prule, ppos, pparent = pitem
					plhs, prhs = prule

					if prhs[ppos:ppos+1] == (lhs,):
						new = (prule,
						       ppos+1,
						       pparent)
						if new not in state:
							state.append(new)
							tree[(new, i)] = [(item, i)]
						else:
							tree[(new, i)].append((item, i))
				continue

			nextSym = rhs[pos]

			#
			#  A -> a . B (predictor)
			#
			if self.rules.has_key(nextSym):
				#
				#  Work on completer step some more; for rules
				#  with empty RHS, the "parent state" is the
				#  current state we're adding Earley items to,
				#  so the Earley items the completer step needs
				#  may not all be present when it runs.
				#
				if needsCompletion.has_key(nextSym):
					new = (rule, pos+1, parent)
					olditem_i = needsCompletion[nextSym]
					if new not in state:
						state.append(new)
						tree[(new, i)] = [olditem_i]
					else:
						tree[(new, i)].append(olditem_i)

				#
				#  Has this been predicted already?
				#
				if predicted.has_key(nextSym):
					continue
				predicted[nextSym] = 1

				ttype = token is not self._EOF and \
					self.typestring(token) or \
					None
				if ttype is not None:
					#
					#  Even smarter predictor, when the
					#  token's type is known.  The code is
					#  grungy, but runs pretty fast.  Three
					#  cases are looked for: rules with
					#  empty RHS; first symbol on RHS is a
					#  terminal; first symbol on RHS is a
					#  nonterminal (and isn't nullable).
					#
					for prule in self.rules[nextSym]:
						new = (prule, 0, i)
						prhs = prule[1]
						if len(prhs) == 0:
							state.append(new)
							continue
						prhs0 = prhs[0]
						if not self.rules.has_key(prhs0):
							if prhs0 != ttype:
								continue
							else:
								state.append(new)
								continue
						first = self.first[prhs0]
						if not first.has_key(None) and \
						   not first.has_key(ttype):
							continue
						state.append(new)
					continue

				for prule in self.rules[nextSym]:
					#
					#  Smarter predictor, as per Grune &
					#  Jacobs' _Parsing Techniques_.  Not
					#  as good as FIRST sets though.
					#
					prhs = prule[1]
					if len(prhs) > 0 and \
					   not self.rules.has_key(prhs[0]) and \
					   token != prhs[0]:
						continue
					state.append((prule, 0, i))

			#
			#  A -> a . c (scanner)
			#
			elif token == nextSym:
				#assert new not in states[i+1]
				states[i+1].append((rule, pos+1, parent))

	def buildTree(self, tokens, tree, root):
		stack = []
		self.buildTree_r(stack, tokens, -1, tree, root)
		return stack[0]

	def buildTree_r(self, stack, tokens, tokpos, tree, root):
		(rule, pos, parent), state = root
		
		while pos > 0:
			want = ((rule, pos, parent), state)
			if not tree.has_key(want):
				#
				#  Since pos > 0, it didn't come from closure,
				#  and if it isn't in tree[], then there must
				#  be a terminal symbol to the left of the dot.
				#  (It must be from a "scanner" step.)
				#
				pos = pos - 1
				state = state - 1
				stack.insert(0, tokens[tokpos])
				tokpos = tokpos - 1
			else:
				#
				#  There's a NT to the left of the dot.
				#  Follow the tree pointer recursively (>1
				#  tree pointers from it indicates ambiguity).
				#  Since the item must have come about from a
				#  "completer" step, the state where the item
				#  came from must be the parent state of the
				#  item the tree pointer points to.
				#
				children = tree[want]
				if len(children) > 1:
					child = self.ambiguity(children)
				else:
					child = children[0]
				
				tokpos = self.buildTree_r(stack,
							  tokens, tokpos,
							  tree, child)
				pos = pos - 1
				(crule, cpos, cparent), cstate = child
				state = cparent
				
		lhs, rhs = rule
		result = self.rule2func[rule](stack[:len(rhs)])
		stack[:len(rhs)] = [result]
		return tokpos

	def ambiguity(self, children):
		#
		#  XXX - problem here and in collectRules() if the same
		#	 rule appears in >1 method.  But in that case the
		#	 user probably gets what they deserve :-)  Also
		#	 undefined results if rules causing the ambiguity
		#	 appear in the same method.
		#
		sortlist = []
		name2index = {}
		for i in range(len(children)):
			((rule, pos, parent), index) = children[i]
			lhs, rhs = rule
			name = self.rule2name[rule]
			sortlist.append((len(rhs), name))
			name2index[name] = i
		sortlist.sort()
		list = map(lambda (a,b): b, sortlist)
		return children[name2index[self.resolve(list)]]

	def resolve(self, list):
		#
		#  Resolve ambiguity in favor of the shortest RHS.
		#  Since we walk the tree from the top down, this
		#  should effectively resolve in favor of a "shift".
		#
		return list[0]

#
#  GenericASTBuilder automagically constructs a concrete/abstract syntax tree
#  for a given input.  The extra argument is a class (not an instance!)
#  which supports the "__setslice__" and "__len__" methods.
#
#  XXX - silently overrides any user code in methods.
#

class GenericASTBuilder(GenericParser):
	def __init__(self, AST, start):
		GenericParser.__init__(self, start)
		self.AST = AST

	def preprocess(self, rule, func):
		rebind = lambda lhs, self=self: \
				lambda args, lhs=lhs, self=self: \
					self.buildASTNode(args, lhs)
		lhs, rhs = rule
		return rule, rebind(lhs)

	def buildASTNode(self, args, lhs):
		children = []
		for arg in args:
			if isinstance(arg, self.AST):
				children.append(arg)
			else:
				children.append(self.terminal(arg))
		return self.nonterminal(lhs, children)

	def terminal(self, token):	return token

	def nonterminal(self, type, args):
		rv = self.AST(type)
		rv[:len(args)] = args
		return rv

#
#  GenericASTTraversal is a Visitor pattern according to Design Patterns.  For
#  each node it attempts to invoke the method n_<node type>, falling
#  back onto the default() method if the n_* can't be found.  The preorder
#  traversal also looks for an exit hook named n_<node type>_exit (no default
#  routine is called if it's not found).  To prematurely halt traversal
#  of a subtree, call the prune() method -- this only makes sense for a
#  preorder traversal.  Node type is determined via the typestring() method.
#

class GenericASTTraversalPruningException:
	pass

class GenericASTTraversal:
	def __init__(self, ast):
		self.ast = ast

	def typestring(self, node):
		return node.type

	def prune(self):
		raise GenericASTTraversalPruningException

	def preorder(self, node=None):
		if node is None:
			node = self.ast

		try:
			name = 'n_' + self.typestring(node)
			if hasattr(self, name):
				func = getattr(self, name)
				func(node)
			else:
				self.default(node)
		except GenericASTTraversalPruningException:
			return

		for kid in node:
			self.preorder(kid)

		name = name + '_exit'
		if hasattr(self, name):
			func = getattr(self, name)
			func(node)

	def postorder(self, node=None):
		if node is None:
			node = self.ast

		for kid in node:
			self.postorder(kid)

		name = 'n_' + self.typestring(node)
		if hasattr(self, name):
			func = getattr(self, name)
			func(node)
		else:
			self.default(node)


	def default(self, node):
		pass

#
#  GenericASTMatcher.  AST nodes must have "__getitem__" and "__cmp__"
#  implemented.
#
#  XXX - makes assumptions about how GenericParser walks the parse tree.
#

class GenericASTMatcher(GenericParser):
	def __init__(self, start, ast):
		GenericParser.__init__(self, start)
		self.ast = ast

	def preprocess(self, rule, func):
		rebind = lambda func, self=self: \
				lambda args, func=func, self=self: \
					self.foundMatch(args, func)
		lhs, rhs = rule
		rhslist = list(rhs)
		rhslist.reverse()

		return (lhs, tuple(rhslist)), rebind(func)

	def foundMatch(self, args, func):
		func(args[-1])
		return args[-1]

	def match_r(self, node):
		self.input.insert(0, node)
		children = 0

		for child in node:
			if children == 0:
				self.input.insert(0, '(')
			children = children + 1
			self.match_r(child)

		if children > 0:
			self.input.insert(0, ')')

	def match(self, ast=None):
		if ast is None:
			ast = self.ast
		self.input = []

		self.match_r(ast)
		self.parse(self.input)

	def resolve(self, list):
		#
		#  Resolve ambiguity in favor of the longest RHS.
		#
		return list[-1]

def _dump(tokens, states):
	for i in range(len(states)):
		print 'state', i
		for (lhs, rhs), pos, parent in states[i]:
			print '\t', lhs, '::=',
			print string.join(rhs[:pos]),
			print '.',
			print string.join(rhs[pos:]),
			print ',', parent
		if i < len(tokens):
			print
			print 'token', str(tokens[i])
			print



#!/usr/bin/env python

#
# Test script for grouch.schema module
#

__revision__ = "$Id: test_schema.py,v 1.11 2001/08/21 22:20:28 gward Exp $"

import sys
from types import LongType, IntType, FileType

from mems.tools.unittest import TestScenario, parse_args, run_scenarios
from cStringIO import StringIO
from grouch.schema import ObjectSchema, ClassDefinition

import fake_module                      # sets up foo.Foo, foo.SubFoo,
                                        # and foo.bar.FooBar modules/classes

tested_modules = [ "grouch.schema" ]


def make_classdef (name, schema, bases=None, attrs=None):
    # name : string
    # bases : [string]
    # attrs : [(string, ValueType)]

    classdef = ClassDefinition(name, schema, bases)
    if attrs:
        for (name, type) in attrs:
            classdef.add_attribute(name, type)
    classdef.finish_definition()
    schema.add_class(classdef)

    return classdef


class SchemaTest (TestScenario):

    def setup (self):
        pass

    def shutdown (self):
        pass

    test_cases = ['create', 'atomic', 'alias', 'classdef', 'gettype',
                  'pickle']

    def check_create (self):
        "create an object schema: 4"
        schema = ObjectSchema()
        self.test_bool("schema.is_atomic_type_name('int')")
        self.test_bool("schema.is_atomic_type_name('function')", want_true=0)
        self.test_val("schema.get_atomic_type_name(LongType)", "long")
        self.test_val("schema.get_atomic_type_object('long')", LongType)

    def check_atomic (self):
        "add/fetch atomic types to/from a schema: 16"

        schema = ObjectSchema()
        int_type = schema.get_type(IntType)
        self.test_bool("int_type.is_atomic_type()")
        self.test_val("str(int_type)", "int")
        self.test_val("schema.get_type('int')", int_type)
        self.test_val("schema.get_type(int_type)", int_type)

        schema.add_atomic_type(sys.stdout)
        self.test_val("schema.get_atomic_type_name(FileType)", "file")
        self.test_val("schema.get_atomic_type_object('file')", FileType)
        self.test_val("schema.get_type('file').type", FileType)
        self.test_bool("schema.is_atomic_type_name('file')")
        self.test_bool("schema.is_atomic_type_name('thingy')", want_true=0)

        schema = ObjectSchema()
        schema.add_atomic_type(sys.stdout, "thingy")
        self.test_val("schema.get_atomic_type_name(FileType)", "thingy")
        self.test_val("schema.get_atomic_type_object('thingy')", FileType)
        self.test_val("schema.get_type('thingy').type", FileType)
        self.test_bool("schema.is_atomic_type_name('thingy')")
        self.test_bool("schema.is_atomic_type_name('file')", want_true=0)

        self.test_exc("schema.add_atomic_type(sys.stdout, 'file')",
                      ValueError)
        self.test_exc("schema.add_atomic_type(self.setup, 'thingy')",
                      ValueError)

    def check_alias (self):
        "adding aliases to a schema: 4"
        schema = ObjectSchema()
        schema.add_alias("real", "int|float|long")
        schema.add_alias("pvalue", "(float,string)")
        self.test_bool("schema.has_alias('real')")
        self.test_bool("schema.get_alias('real').is_union_type()")

        schema.add_alias("range", "(real,real)")
        self.test_bool("schema.has_alias('range')")
        self.test_val("schema.get_alias('range').container_type", "tuple")

    def check_classdef (self):
        "add/fetch class definitions to/from a schema: 5"
        import foo
        schema = ObjectSchema()

        self.test_val("schema.get_class_definition('foo.Foo')", None)
        self.test_val("schema.get_class_definition(foo.Foo)", None)

        Foo_def = make_classdef("foo.Foo", schema)
        self.test_val("schema.get_class_definition('foo.Foo')", Foo_def)
        self.test_val("schema.get_class_definition(foo.Foo)", Foo_def)
        self.test_exc("schema.add_class(Foo_def)", RuntimeError)

    def check_gettype (self):
        "exercise the wildly overloaded 'get_type()' method: 9"
        from types import FunctionType  # not an atomic type!
        import foo

        schema = fake_module.make_schema() # so we have some classes defined
        schema.add_alias("Foo", "foo.Foo")

        self.test_val("schema.get_type(IntType).type", IntType)
        self.test_exc("schema.get_type(FunctionType)", ValueError)

        self.test_val("schema.get_type('int').type", IntType)
        self.test_val("schema.get_type('foo.Foo').klass_name","foo.Foo")
        self.test_val("schema.get_type('Foo').alias_type.klass_name","foo.Foo")
        self.test_exc("schema.get_type('blah')", ValueError)
        
        self.test_val("schema.get_type(foo.Foo).klass_name", "foo.Foo")
        int_type = schema.make_atomic_type('int')
        self.test_val("schema.get_type(int_type)", int_type, match_ident=1)

        self.test_exc("schema.get_type(1)", TypeError)

    def check_pickle (self):
        "pickle an object schema: 8"
        from cPickle import loads, dumps
        orig = ObjectSchema()
        copy = loads(dumps(orig))

        # Redo the tests for a just-created schema on the recreated one
        self.test_bool("copy.is_atomic_type_name('int')")
        self.test_bool("copy.is_atomic_type_name('function')", want_true=0)
        self.test_val("copy.get_atomic_type_name(LongType)", "long")
        self.test_val("copy.get_atomic_type_object('long')", LongType)

        inttype = copy.get_type(IntType)
        longtype = copy.get_type(LongType)
        self.test_bool("inttype.schema is longtype.schema")

        # Make sure we can pickle a schema with scanner and parser objects
        # (SPARK classes don't seem to be pickleable).
        orig.parse_type("[int]")
        oparser = orig.parser
        self.test_stmt("dumps(orig)")
        self.test_bool("orig.parser is oparser")
        copy = loads(dumps(orig))
        self.test_stmt("copy.parse_type('[int]')")

# class SchemaTest


class ClassDefTest (TestScenario):
    
    def setup (self):
        self.schema = ObjectSchema()
        self.int_type = self.schema.get_type('int')
        self.float_type = self.schema.get_type('float')

    def shutdown (self):
        pass


    test_cases = ['create', 'attrs', 'write']

    def make_classdef (self, name, bases=None, attrs=None):
        return make_classdef(name, self.schema, bases, attrs)


    def check_create (self):
        "create a class definition: 4"
        # Hmmm, this'll be in trouble if ClassDefinition ever starts
        # to enforce that the named class must exist...
        cdef = ClassDefinition("Foo", self.schema)
        self.test_val("cdef.name", "Foo")
        self.test_val("cdef.bases", [])
        self.test_val("cdef.num_attributes()", 0)

        cdef = ClassDefinition("SubFoo", self.schema, ["Foo"])
        self.test_val("cdef.bases", ["Foo"])

    def check_attrs (self):
        "add/fetch attribute definitions: 18"
        Foo_def = self.make_classdef("Foo",
                                     attrs=[("a", self.int_type),
                                            ("b", self.float_type)])
        self.test_val("Foo_def.num_attributes()", 2)
        self.test_val("Foo_def.get_attribute_type('a')", self.int_type)

        # Should make this work eventually.
        self.test_exc("Foo_def.add_attribute('a', 'int')", TypeError)
        # But not this!
        self.test_exc("Foo_def.add_attribute('a', 42)", TypeError)

        Bar_def = self.make_classdef("Bar", attrs=[("c", self.int_type)])
        SubFoo_def = self.make_classdef("SubFoo",
                                        bases=["Foo", "Bar"],
                                        attrs=[("d", self.float_type),
                                               ("e", self.int_type)])
        self.test_val("SubFoo_def._find_bases()",
                      [Foo_def, Bar_def, SubFoo_def])
        self.test_val("len(SubFoo_def.attrs)", 2)
        self.test_val("len(SubFoo_def.attr_types)", 2)
        self.test_val("SubFoo_def.num_attributes()", 5)
        self.test_val("SubFoo_def.get_attribute_type('a')", self.int_type)
        self.test_val("SubFoo_def.get_attribute_type('c')", self.int_type)
        self.test_val("SubFoo_def.get_attribute_type('d')", self.float_type)

        SubSubFoo_def = self.make_classdef("SubSubFoo",
                                           bases=["SubFoo"])
        self.test_val("SubSubFoo_def._find_bases()",
                      [Foo_def, Bar_def, SubFoo_def, SubSubFoo_def])
        self.test_val("len(SubSubFoo_def.attrs)", 0)
        self.test_val("len(SubSubFoo_def.attr_types)", 0)
        self.test_val("SubSubFoo_def.num_attributes()", 5)
        self.test_val("SubSubFoo_def.get_attribute_type('a')", self.int_type)
        self.test_val("SubSubFoo_def.get_attribute_type('c')", self.int_type)
        self.test_val("SubSubFoo_def.get_attribute_type('d')", self.float_type)

    def check_write (self):
        "write class definition: 3"
        
        cdef = self.make_classdef("Bar")
        expect = """\
class Bar:
    pass
"""
        f = StringIO() ; cdef.write(f)
        self.test_val("f.getvalue()", expect)

        cdef = self.make_classdef("Foo",
                                  attrs=[("a", self.int_type),
                                         ("b", self.float_type)])
        expect = """\
class Foo:
    a : int
    b : float
"""
        f = StringIO() ; cdef.write(f)
        self.test_val("f.getvalue()", expect)

        cdef = self.make_classdef("SubFoo", bases=["Foo"])
        expect = """\
class SubFoo (Foo):
    pass
"""
        f = StringIO() ; cdef.write(f)
        self.test_val("f.getvalue()", expect)

# class ClassDefTest


if __name__ == "__main__":
    (scenarios, options) = parse_args()
    run_scenarios (scenarios, options)



"""grouch.util

Miscellaneous utility functions.
"""

# created 2000/12/13, GPW
#
# Copyright (c) 2001 CNRI.  All Rights Reserved.

__revision__ = "$Id: util.py,v 1.9 2001/09/05 14:21:23 gward Exp $"

import sys
import __builtin__
from types import TypeType, ClassType, InstanceType


def is_class_object (klass):
    """Return true if 'klass' is a class object (either a regular Python
    class or an ExtensionClass).
    """
    # The first clause handles regular classes quickly; the second
    # is needed to accomodate ExtensionClasses
    return type(klass) is ClassType or hasattr(klass, '__bases__')

def is_instance_object (object):
    """Return true if 'object' is an instance of some class (either a
    regular Python class or an ExtensionClass).
    """
    return ((type(object) is InstanceType or
             (hasattr(object, '__class__') and
              not hasattr(object, '__bases__'))) and
            hasattr(object, '__dict__'))

def get_class_object (name):
    lastdot = name.rfind(".")
    if lastdot == -1:                   # no dots -- look in main module
        module_name = '__main__'
        klass_name = name
        module = sys.modules[module_name]
    else:
        module_name = name[:lastdot]
        klass_name = name[lastdot+1:]
        try:
            __import__(module_name)
        except ImportError, exc:
            raise ValueError, \
                  "error finding class %s: %s" % (name, exc)
        module = sys.modules[module_name]

    if not hasattr(module, klass_name):
        raise ValueError, \
              "no such class %s in module %s" % (`klass_name`, module_name)
    else:
        klass = getattr(module, klass_name)
        if is_class_object(klass):
            return klass
        else:
            raise ValueError, \
                  "%s in module %s: not a class" % (`klass_name`, module_name)


def get_full_classname (klass):
    classname = klass.__name__
    modname = klass.__module__
    if modname == "__main__":
        return classname
    else:
        return "%s.%s" % (modname, classname)


# Supposedly, these two won't be necessary with Python 2.1, since its
# 'issubclass()' and 'isinstance()' will finally be able to deal with
# extension classes completely in 2.1.  (It's close in 1.6 and 2.0, but not
# quite there yet.)  I have not yet verified this!

def issubclass (class1, class2):
    """A version of 'issubclass()' that works with extension classes
    as well as regular Python classes.
    """

    # Both class objects are regular Python classes, so use the
    # built-in 'issubclass()'.
    if type(class1) is ClassType and type(class2) is ClassType:
        return __builtin__.issubclass(class1, class2)

    # Both so-called class objects have a '__bases__' attribute: ie.,
    # they aren't regular Python classes, but they sure look like them.
    # Assume they are extension classes and reimplement what the builtin
    # 'issubclass()' does behind the scenes.
    elif hasattr(class1, '__bases__') and hasattr(class2, '__bases__'):
        # XXX it appears that "ec.__class__ is type(ec)" for an
        # extension class 'ec': could we/should we use this as an
        # additional check for extension classes?

        # Breadth-first traversal of class1's superclass tree.  Order
        # doesn't matter because we're just looking for a "yes/no"
        # answer from the tree; if we were trying to resolve a name,
        # order would be important!
        stack = [class1]
        while stack:
            if stack[0] is class2:
                return 1
            stack.extend(list(stack[0].__bases__))
            del stack[0]
        else:
            return 0

    # Not a regular class, not an extension class: blow up for consistency
    # with builtin 'issubclass()"
    else:
        raise TypeError, "arguments must be class or ExtensionClass objects"

# issubclass ()

def isinstance (object, klass):
    """A version of 'isinstance()' that works with extension classes
    as well as regular Python classes."""

    if type(klass) is TypeType:
        return __builtin__.isinstance(object, klass)
    elif hasattr(object, '__class__'):
        return issubclass(object.__class__, klass)
    else:
        return 0



from thing import Thing

class Animal (Thing):
    """An animal, ie. a thing with multiple legs and possibly fur.

    Instance attributes:
      num_legs : int
        the number of legs this animal has
      furry : boolean
        whether this animal is furry or not

    Outsiders should use 'get_num_legs()' and 'is_furry()' to access
    these attributes.
    """

    def __init__ (self, name):
        Thing.__init__(self, name)
        self.num_legs = None
        self.furry = None

    def set_num_legs (self, n):
        self.num_legs = n

    def set_furry (self, furry):
        self.furry = furry

    def get_num_legs (self):
        return self.num_legs

    def is_furry (self):
        return self.furry


class Mammal (Animal):
    """A warm-blooded animal that is born as a baby (not an egg).

    Instance attributes: none
    """

    def __init__ (self, name):
        Animal.__init__(self, name)
        self.furry = 1                  # mammals usually are



"""grouch.script_util

Code common to the Grouch scripts.
"""

# Copyright (c) 2001 CNRI.  All Rights Reserved.

__revision__ = "$Id: script_util.py,v 1.3 2001/09/05 14:21:23 gward Exp $"

import sys

def writenow (msg):
    sys.stdout.write(msg)
    sys.stdout.flush()

def announce (msg, threshold=1):
    import __main__
    if __main__.VERBOSITY >= threshold:
        writenow(msg)

def error (msg):
    sys.stderr.write("error: " + msg + "\n")

def warn (msg):
    sys.stderr.write("warning: " + msg + "\n")



"""grouch.valuetype

Provides the ValueType class and its subclasses, which together
represent the Python type system as interpreted by Grouch.
"""

# created 2000/12/13, GPW (extracted from mems.tools.schema module)
# rewritten 2001/02/02, GPW (refactored overgrown ValueType class, and
#                            extracted parsing code to parsetype.py)
#
# Copyright (c) 2001 CNRI.  All Rights Reserved.

__revision__ = "$Id: valuetype.py,v 1.35 2001/09/05 14:21:23 gward Exp $"

import string
import new
import types

from grouch.util import is_class_object, is_instance_object, get_full_classname


# XXX Type-checking issues yet to be resolved
#
#  * if an "any" type has 'allow_none' false, does it allow None?
#    eg. if 'a' is declared "a : any notnone", can 'a' be None?
#  * ditto for boolean types:
#    if 'f' is decalred "f : boolean notnone", can 'f' be None?
#  * what does an empty/uninitialized union type mean?  no values at
#    all are valid?  perhaps only None should be allowed?


class ValueType:
    """
    Instance attributes:
      schema : ObjectSchema
        the object schema (atomic types, aliases, and class definitions)
        relative to which we should interpret this type
      allow_none : boolean = true
        are values allowed to be None?
    """

    # class attribute 'metatype' must be defined by each subclass
    metatype = None


    def __init__ (self, schema):
        self.schema = schema
        self.allow_none = 1

        assert self.__class__ is not ValueType,\
               "ValueType is an abstract class"
        assert self.metatype is not None,\
               "metatype must be set by subclass"


    # Subclasses must provide '__str__()'

    def __repr__ (self):
        return "<%s at %x: %s>" % (self.__class__.__name__, id(self), self)

    # Pickle support: every class that might ultimately contain some
    # AtomicType instance needs to implement this; it should call
    # '_fixup_atomic_types()' on all referenced ValueType instances.  See
    # the comment in AtomicType for details on pickle support.
    def _fixup_atomic_types (self, memo):
        pass

    # -- Accessors/modifiers -------------------------------------------

    def get_metatype (self):
        return self.metatype            # really a class attribute!


    # -- Predicates ----------------------------------------------------
    # XXX might be cleaner (and more extensible) to just "return 0"
    # everywhere here, and require subclasses to override and
    # "return 1" where applicable

    def is_atomic_type (self):
        return self.metatype == "atomic"

    def is_instance_type (self):
        return self.metatype in ("instance", "instance-container")

    def is_plain_instance_type (self):
        return self.metatype == "instance"

    def is_container_type (self):
        return self.metatype in ("container", "instance-container")

    def is_plain_container_type (self):
        return self.metatype == "container"

    def is_instance_container_type (self):
        return self.metatype == "instance-container"

    def is_union_type (self):
        return self.metatype == "union"

    def is_any_type (self):
        return self.metatype == "any"

    def is_boolean_type (self):
        return self.metatype == "boolean"

    def is_alias_type (self):
        return self.metatype == "alias"


    # -- Type-checking -------------------------------------------------

    def check_value (self, value, context):
        raise NotImplementedError, "subclasses must implement"

    def check_none_ok (self, value):
        return (value is None and self.allow_none)
    
    def check_unexpected_instance (self, value, context, allow_foreign=0):
        if not is_instance_object(value):
            return 1
        try:
            itype = self.schema.make_instance_type(value.__class__)
        except ValueError, msg:
            # XXX ugh, should probably have some custom exceptions
            msg = str(msg)              # just in case
            if not msg.startswith("class not in schema"):
                raise

            if allow_foreign:
                return 1
            else:
                context.add_error("found instance of class not in schema: %s" %
                                  get_full_classname(value.__class__))
                return 0
        else:
            return itype.check_value(value, context)

    def count_object (self, context):
        context.count_object(self)

    def uncount_object (self, context):
        context.uncount_object(self)


class AtomicType (ValueType):
    """
    Instance attributes:
      type : type
        the type object that values must match
      type_name : string
        the name of the atomic type
    """

    metatype = "atomic"

    def __init__ (self, schema):
        ValueType.__init__(self, schema)
        self.type = None
        self.type_name = None

    def __str__ (self):
        return self.type_name or "*no type*"

    # -- Pickle support ------------------------------------------------
    # (needed because you can't pickle type objects)

    # Pickling ObjectSchema instances (and their associated
    # ClassDefinitions and ValueTypes) is rather tricky because you can't
    # pickle type objects and because of the incestuous web of
    # interdependencies between the three classes.  We have to purge type
    # objects at pickle time, and then somehow reconstitute them at
    # unpickle time.
    #
    # Things gets tricky under the following circumstances:
    #  - the root of the pickle is an ObjectSchema
    #  - we're somewhere deep in the guts of the ObjectSchema and
    #    stumble across a AtomicType that wants to be unpickled
    # 
    # At this point, we need to reconstruct the AtomicType's 'type'
    # attribute, which it needs for type-checking.  We need to query the
    # ObjectSchema's 'atomic_types' dictionary to turn a type name (which
    # is what we pickled in lieu of the unpickleable type object) into its
    # type object.  However, since we're in the middle of unpickling the
    # ObjectSchema, its 'atomic_types' dictionary doesn't exist yet.  (See
    # ObjectSchema's '__{get,set}state__()'.)  We can't finish unpickling
    # the AtomicType because the ObjectSchema isn't finished unpickling,
    # and the ObjectSchema won't be usable until all of its AtomicTypes are
    # finished unpickling: Catch-22!
    #
    # The workaround is to add a "fixup" pass after unpickling is
    # all-but-finished.  (It's done as the last part of ObjectSchema's
    # '__setstate__()'.)  This gives control back to every AtomicType in
    # the schema just long enough for it to convert a type name to a type
    # object, using the now-fully-unpickled ObjectSchema.  The catch is
    # that almost every ValueType subclass needs to implement the fixup
    # method, just so we can be sure to hit every atomic type in the
    # schema.
    #
    # The catch is that pickling AtomicType instances on their own doesn't
    # work very well.  Executive summary: don't do it.  If you must, then
    # be sure you understand what you are doing, and call
    # '_fixup_atomic_types()' afterwards.
    #
    # Obviously, if the semantics of Python's pickle (specifically, the
    # order in which objects are unpickled) ever change significantly, we
    # could be in deep doo-doo here.

    def __getstate__ (self):
        if self.type is None:
            return self.__dict__
        else:
            from copy import copy
            dict = copy(self.__dict__)
            del dict['type']
            return dict

    def __setstate__ (self, dict):
        from grouch.schema import ObjectSchema
        self.__dict__.update(dict)

    def _fixup_atomic_types (self, memo):
        if not hasattr(self, 'type'):
            self.type = self.schema.get_atomic_type_object(self.type_name)


    def set_type (self, tipe):
        tipe = self.schema.get_atomic_type_object(tipe)
        self.type = tipe
        self.type_name = self.schema.get_atomic_type_name(tipe)

    def get_type (self):
        return self.type


    def check_value (self, value, context):
        self.count_object(context)
        vtype = type(value)
        if vtype is self.type or self.check_none_ok(value):
            return 1
        else:
            context.add_error("expected %s, got %s" %
                              (str(self), describe_value(value)))
            self.check_unexpected_instance(value, context)
            return 0

    def count_object (self, context):
        context.count_object(self, self.type_name)

    def uncount_object (self, context):
        context.uncount_object(self, self.type_name)

# class AtomicType


# XXX name collision w/ standard 'types' modules!
class InstanceType (ValueType):
    """
    Instance attributes:
      klass_name : string
        the (fully-qualified) name of the class
      klass_definition : ClassDefinition
        the definition of this class in the object schema (defines
        what attributes are allowed/expected and what their
        types are)
      klass : class
        a class object; values must be instances of this class

    XXX klass and klass_definition might be obsolete; they are now
    assigned only when needed, ie. in check_value()
    """

    metatype = "instance"

    def __init__ (self, schema):
        ValueType.__init__(self, schema)
        self.klass_name = None
        self.klass_definition = None
        self.klass = None

    def _fixup_atomic_types (self, memo):
        if self.klass_definition:
            self.klass_definition._fixup_atomic_types(memo)

    def __str__ (self):
        return self.klass_name or "*no class*"

    def set_class_name (self, klass_name):
        if type(klass_name) is not types.StringType:
            raise TypeError, "not a string: %r" % klass_name
        self.klass_name = klass_name

    def get_class_name (self):
        return self.klass_name


    # Helper methods for 'check_value()'

    def _find_classdef (self, value, context):
        klass = value.__class__
        if klass is self.klass:
            klass_def = self.klass_definition
        else:
            klass_def = self.schema.get_class_definition(klass)
            if not klass_def:
                context.add_error("expected instance of %s, "
                                  "got instance of unknown subclass %s"
                                  % (self.klass,
                                     get_full_classname(klass)))

        return klass_def
        
    def _check_attrs (self, value, klass_def, context):

        #print "checking " + repr(value)

        # First, make sure there are no unexpected attributes.  (Doing this
        # first -- in the instance dictionary's hash order -- means the
        # *real* type-checking loop, below, is done according to the order
        # of attributes in the class definition, ie. in a predictable,
        # canonical order.  This is a good thing, if only because it makes
        # writing tests easier.)

        if context.mode == "zodb":
            # prod ZODB into loading the object before relying on
            # the value of dir()
            dir(value)

        all_ok = 1
        actual_attrs = dir(value)
        for actual_attr in actual_attrs:
            if context.mode == "zodb" and actual_attr.startswith("_v_"):
                continue

            if not klass_def.get_attribute_type(actual_attr):
                val = getattr(value, actual_attr)
                context.add_error("found unexpected attribute '%s': %s" %
                                  (actual_attr, describe_value(val)))

                context.push_context('attr', actual_attr)
                self.check_unexpected_instance(val, context)
                context.pop_context()

                all_ok = 0

        # Now the real type-checking loop, in canonical (not hash)
        # order.  By-the-by, we also check for missing attributes here.
        for expected_attr in klass_def.all_attrs:
            if expected_attr in actual_attrs:
                val = getattr(value, expected_attr)
                expected_type = klass_def.get_attribute_type(expected_attr)
                context.push_context('attr', expected_attr)
                ok = expected_type.check_value(val, context)
                if not ok:
                    all_ok = 0

                context.pop_context()
            else:
                context.add_error("expected attribute '%s' not found" %
                                  expected_attr)
                all_ok = 0

        return all_ok

    def check_value (self, value, context):
        # get the special isinstance() that knows about ExtensionClass
        from grouch.util import isinstance, get_class_object

        # XXX from Jim Fulton on zodb-dev@zope.org, about how to
        # reduce memory use of the type-checking script:
        # 
        # You could modify your script to your script to "ghostify" the
        # objects, by calling _p_deactivate on them after you use them
        # to prevent the memory bloat.  This is what the Zope "find"
        # facility does when searching for objects.
        #
        # I think this is the place to do it, but this method would need
        # a bit of rearrangement.

        if self.klass is None:
            self.klass = get_class_object(self.klass_name)
        if self.klass_definition is None:
            self.klass_definition = \
                self.schema.get_class_definition(self.klass_name)

        self.count_object(context)
        if isinstance(value, self.klass):
            klass_def = self._find_classdef(value, context)
            if not klass_def:
                return 0
            
            if context.check_seen(value):
                # No new errors to add here, so act as though the guts of
                # the object are fine.
                self.uncount_object(context)
                return 1
            
                # XXX maybe the context should keep track of which
                # instances/containers had errors in them, so we can say
                # "second visit to object with errors" -- hmm, that'll only
                # work with instances, because a container might visited
                # twice with different declarations in force.

            #print "checking", repr(value)
            return self._check_attrs(value, klass_def, context)

        elif self.check_none_ok(value):
            return 1

        else:
            context.add_error("expected instance of %s, got %s" %
                              (self.klass, describe_value(value)))
            self.check_unexpected_instance(value, context)
            return 0

        # if/elif/else

    # check_value ()

    def count_object (self, context):
        context.count_object(self, self.klass_name)

    def uncount_object (self, context):
        context.uncount_object(self, self.klass_name)

# class InstanceType


class ContainerType (ValueType):
    """
    Instance attributes:
      allow_empty : boolean = true
        only for container and instance-container types: are
        values allowed to be empty containers?
    """

    metatype = "container"
    container_type = None               # subclasses must set

    def __init__ (self, schema):
        assert self.__class__ is not ContainerType,\
               "ContainerType is an abstract class"

        ValueType.__init__(self, schema)
        self.allow_empty = 1


    # -- Type-checking methods -----------------------------------------

    container_type_object = {
        "list" : types.ListType,
        "tuple" : types.TupleType,
        "dictionary" : types.DictionaryType,
        }

    def check_value (self, value, context):
        self.count_object(context)
        if type(value) is self.container_type_object[self.container_type]:
            if len(value) == 0 and self.container_type != 'tuple':
                if not self.allow_empty:
                    context.add_error("empty %s not allowed" %
                                      self.container_type)
                    return 0
                else:
                    return 1

            return self.check_contents(value, context)
        elif self.check_none_ok(value):
            return 1
        else:
            context.add_error("expected %s, got %s" %
                              (self.container_type, describe_value(value)))
            self.check_unexpected_instance(value, context)
            return 0

    def count_object (self, context):
        context.count_object(self, self.container_type)

    def uncount_object (self, context):
        context.uncount_object(self, self.container_type)

# class ContainerType


class ListType (ContainerType):
    """
    Instance attributes:
      element_type : ValueType
        the type that all elements of the list must be
    """

    container_type = "list"

    def __init__ (self, schema):
        ContainerType.__init__(self, schema)
        self.element_type = None

    def _fixup_atomic_types (self, memo):
        self.element_type._fixup_atomic_types(memo)

    def __str__ (self):
        return "[%s]" % self.element_type

    def set_element_type (self, element_type):
        self.element_type = self.schema.get_type(element_type)

    set_list_type = set_element_type    # backwards compat.

    def get_element_type (self):
        return self.element_type


    # -- Type-checking methods -----------------------------------------

    def check_contents (self, value, context):
        i = 0
        all_ok = 1
        for item in value:
            context.push_context('item', i)
            ok = self.element_type.check_value(item, context)
            if not ok:
                all_ok = 0
            context.pop_context()
            i = i + 1
            
        return all_ok


class TupleType (ContainerType):
    """
    Instance attributes:
      element_types : [ValueType]
        the type of each "slot" in the tuple (values of this type
        must be tuples of the same length as 'element_types', and
        each element of the tuple must match the corresponding
        type in 'element_types'
      extended : boolean = false
        if true, this type is for "extended" tuples
    """

    container_type = "tuple"

    def __init__ (self, schema):
        ContainerType.__init__(self, schema)
        self.element_types = []
        self.extended = 0

    def _fixup_atomic_types (self, memo):
        for subtype in self.element_types:
            subtype._fixup_atomic_types(memo)

    def __str__ (self):
        types = self.element_types
        if self.extended:
            if len(types) == 1:
                return "(%s*,)" % types[0]
            else:
                types = map(str, types)
                return ("(%s, %s*)" %
                        (string.join(types[0:-1], ", "), types[-1]))

        else:                           # regular tuple type
            if len(types) == 1:
                return "(%s,)" % types[0]
            else:
                types = map(str, types)
                return "(%s)" % string.join(types, ", ")

    def set_element_types (self, element_types):
        if type(element_types) not in (types.TupleType, types.ListType):
            raise TypeError, \
                  "not a sequence of type objects: %s" % `element_types`
        self.element_types = map(self.schema.get_type, element_types)

    def set_extended (self, extended):
        self.extended = extended

    set_tuple_type = set_element_types  # backwards compat.

    def get_element_types (self):
        return self.element_types

    def is_extended (self):
        return self.extended

    def check_contents (self, value, context):
        types = self.element_types
        ext = self.extended

        # First, just check the length; if len(element_types) == N:
        #  - regular tuples must be of length N
        #  - extended tuples must be of length >= N-1
        #    (last item repeats 0 .. * times)

        if (not ext and len(value) != len(types)):
            context.add_error("expected tuple of length %d, "
                              "got one of length %d" %
                              (len(types), len(value)))
            all_ok = 0
        elif (ext and len(value) < len(types)-1):
            context.add_error("expected tuple of at least length %d, "
                              "got one of length %d" %
                              (len(types)-1, len(value)))
            all_ok = 0
        else:
            all_ok = 1

        # Length is OK, so now check the actual elements
        if ext:                         # check the whole tuple
            n = len(value)
        else:                           # check only what we have types for
            n = min(len(value), len(types))

        for i in range(n):
            context.push_context('item', i)
            if ext and i >= len(types):
                cur_type = types[-1]
            else:
                cur_type = types[i]

            if not cur_type.check_value(value[i], context):
                all_ok = 0
            context.pop_context()

        return all_ok


class DictionaryType (ContainerType):
    """
    Instance attributes:
      key_type : ValueType
      value_type : ValueType

    """

    container_type = "dictionary"

    def __init__ (self, schema):
        ContainerType.__init__(self, schema)
        self.key_type = None
        self.value_type = None

    def _fixup_atomic_types (self, memo):
        self.key_type._fixup_atomic_types(memo)
        self.value_type._fixup_atomic_types(memo)

    def __str__ (self):
        return "{ %s : %s }" % (self.key_type, self.value_type)

    def set_key_type (self, key_type):
        self.key_type = self.schema.get_type(key_type)

    def set_value_type (self, value_type):
        self.value_type = self.schema.get_type(value_type)

    def set_item_types (self, key_type, value_type):
        self.key_type = self.schema.get_type(key_type)
        self.value_type = self.schema.get_type(value_type)

    set_dictionary_type = set_item_types # backwards compat.


    def get_key_type (self):
        return self.key_type

    def get_value_type (self):
        return self.value_type

    def get_item_types (self):
        return (self.key_type, self.value_type)


    def check_contents (self, value, context):
        all_ok = 1
        for key in value.keys():
            context.push_context('dkey', key)
            if not self.key_type.check_value(key, context):
                all_ok = 0

            context.replace_context('dval', key)
            val = value[key]
            if not self.value_type.check_value(val, context):
                all_ok = 0

            context.pop_context()

        return all_ok


class InstanceContainerType (InstanceType):
    """
    Instance attributes:
      container_type : ContainerType
    """

    metatype = "instance-container"

    def __init__ (self, schema):
        InstanceType.__init__(self, schema)
        self.container_type = None

    def _fixup_atomic_types (self, memo):
        InstanceType._fixup_atomic_types(self, memo)
        self.container_type._fixup_atomic_types(memo)

    def __str__ (self):
        return "%s %s" % (self.klass_name, self.container_type)

    def set_container_type (self, container_type):
        if not isinstance(container_type, ContainerType):
            raise TypeError, \
                  "not a ContainerType instance: %s" % `container_type`
        self.container_type = container_type

    def get_container_type (self):
        return self.container_type

    def get_element_type (self):
        if not isinstance(self.container_type, ListType):
            raise RuntimeError, "not a list type"
        return self.container_type.get_element_type()

    def get_element_types (self):
        if not isinstance(self.container_type, TupleType):
            raise RuntimeError, "not a tuple type"
        return self.container_type.get_element_types()

    def get_item_types (self):
        if not isinstance(self.container_type, DictionaryType):
            raise RuntimeError, "not a dictionary type"
        return self.container_type.get_item_types()


    def check_value (self, value, context):
        cont_ok = self.container_type.check_contents(value, context)
        inst_ok = InstanceType.check_value(self, value, context)
        return inst_ok and cont_ok


class UnionType (ValueType):
    """
    Instance attributes:
      union_types : [ValueType]
        a set of types; values must match one of them
    """

    metatype = "union"

    def __init__ (self, schema):
        ValueType.__init__(self, schema)
        self.union_types = None

    def _fixup_atomic_types (self, memo):
        for subtype in self.union_types:
            subtype._fixup_atomic_types(memo)

    def __str__ (self):
        types = map(str, self.union_types)
        return string.join(types, " | ")

    def set_union_types (self, union_types):
        if type(union_types) is not types.ListType:
            raise TypeError, "not a list of type objects: %s" % `union_types`
        union_types = map(self.schema.get_type, union_types)
        self.union_types = union_types

    def get_union_types (self):
        return self.union_types

    def check_value (self, value, context):

        self.count_object(context)

        # Temporarily turn off error reporting if it's on; we want to
        # replace any errors from sub-types with our own error message.
        save_report = context.report_errors
        context.report_errors = 0

        types = self.union_types
        for subtype in types:
            ok = subtype.check_value(value, context)
            subtype.uncount_object(context)
            if ok:
                context.report_errors = save_report
                return 1
            else:
                # Forget about this error message -- if *all* checks fail,
                # then we'll generate a new message; if *any* pass, then
                # there's no error.
                context.forget_errors(1)
        else:
            context.report_errors = save_report

            if len(types) == 0:
                prefix = "no valid values for empty union type"
            else:
                # Ugh, can forget about I18N of this code...
                tnames = map(str, types)
                if len(types) == 1:
                    tlist = tnames[0]
                elif len(types) == 2:
                    tlist = "%s or %s" % tuple(tnames)
                else:
                    tlist = (string.join(tnames[0:-1], ", ") +
                             ", or " + tnames[-1])
                prefix = "expected " + tlist

            context.add_error(prefix + "; got %s" % describe_value(value))
            self.check_unexpected_instance(value, context)
            return 0

    # check_value ()

    def count_object (self, context):
        context.count_object(self, str(self))

    def uncount_object (self, context):
        context.uncount_object(self, str(self))

# class UnionType


class AnyType (ValueType):
    """
    Instance attributes: none
      allow_any_instance : boolean
        if true (the default), class instances of any class are allowed;
        otherwise, instances must be of classes known to the containing
        object schema.
    """

    metatype = "any"

    def __init__ (self, schema):
        ValueType.__init__(self, schema)
        self.allow_any_instance = 1

    def __str__ (self):
        return "any"

    def set_allow_any_instance (self, allow_any_instance):
        self.allow_any_instance = allow_any_instance

    def check_value (self, value, context):
        self.count_object(context)
        if value is None and not self.allow_none:
            context.add_error("expected anything except None")
            return 0
        else:
            return self.check_unexpected_instance(
                value, context, allow_foreign=self.allow_any_instance)


class BooleanType (ValueType):
    """
    Instance attributes: none
    """

    metatype = "boolean"

    def __init__ (self, schema):
        ValueType.__init__(self, schema)

    def __str__ (self):
        return "boolean"

    def check_value (self, value, context):
        self.count_object(context)
        if value in (0, 1, None):
            return 1
        else:
            context.add_error("expected a boolean, got %s" %
                              describe_value(value))
            self.check_unexpected_instance(value, context)
            return 0

class AliasType (ValueType):
    """
    Instance attributes:
      alias_name : string
        only for alias types: the name of this type (the alias)
      alias_type : ValueType
        the aliased type (what 'alias_name' stands for)
    """

    metatype = "alias"

    def __init__ (self, schema):
        ValueType.__init__(self, schema)
        self.alias_name = None
        self.alias_type = None

    def _fixup_atomic_types (self, memo):
        self.alias_type._fixup_atomic_types(memo)

    def __str__ (self):
        return self.alias_name

    def set_alias_type (self, alias_name, alias_type):
        self.alias_name = alias_name
        self.alias_type = alias_type

    def get_alias_type (self):
        return self.alias_type

    def check_value (self, value, context):
        self.count_object(context)
        return self.alias_type.check_value(value, context)

    def count_object (self, context):
        context.count_object(self, self.alias_name)

    def uncount_object (self, context):
        context.uncount_object(self, self.alias_name)

# -- Utility functions -------------------------------------------------

def describe_value (value):
    if is_instance_object(value):
        # It's an instance of either a vanilla Python class or an
        # ExtensionClass.
        return "instance of %s" % get_full_classname(value.__class__)
    elif value is None:
        return "None"
    else:
        if (type(value) in (types.ListType,
                            types.TupleType,
                            types.DictionaryType)
            and len(value) > 10):
            str_value = "... elided ..."
        elif type(value) is types.StringType and len(value) > 100:
            str_value = value[0:20] + " ... " + value[-20:]
        else:
            str_value = repr(value)
            
        return "%s (%s)" % (type(value).__name__, str_value)



#!/usr/bin/env python

#
# Test script for grouch.parsetype module
#

__revision__ = "$Id: test_parsetype.py,v 1.14 2001/08/21 22:20:28 gward Exp $"

import sys
from types import IntType, StringType, FloatType
from mems.tools.unittest import TestScenario, parse_args, run_scenarios

from grouch.schema import ObjectSchema
from grouch.valuetype import ContainerType, UnionType
from grouch.type_parser import TypeScanner, TypeParser

tested_modules = [ "grouch.parser" ]

# Set up a fake module hierarchy:
#   module 'foo' provides class Foo (and sub-module bar)
#   module 'foo.bar' provides class FooBar
#   module '__main__' provides class Plain
import fake_module

_scanner = TypeScanner()
_parser = TypeParser()

# Emulate the old standalone 'parse_type()' function.
def parse_type (typestr, schema):
    global _scanner, _parser
    _parser.set_schema(schema)
    return _parser.parse(_scanner.tokenize(typestr))


class ParseTypeTest (TestScenario):

    def setup (self):
        self.schema = fake_module.make_schema()

    def shutdown (self):
        del self.schema


    test_cases = ['atomic',
                  'misc',
                  'instance',
                  'list',
                  'tuple',
                  'dict',
                  'instcont',
                  'recursive_1',
                  'recursive_2',
                  'errors',
                  'alias']


    def check_atomic (self):
        "parse atomic and 'any' types: 6"
        t = parse_type('int', self.schema)
        self.test_val("str(t)", "int")
        self.test_bool("t.is_atomic_type()")
        self.test_val("t.type", IntType)

        self.test_exc("parse_type('list', self.schema)", ValueError)
        self.test_exc("parse_type('foo', self.schema)", ValueError)
        self.test_exc("parse_type(42, self.schema)", TypeError)

    def check_misc (self):
        "parse miscellaneous types (any, boolean): 7"
        t = parse_type('any', self.schema)
        self.test_bool("t.is_any_type()")

        t = parse_type('boolean', self.schema)
        self.test_bool("t.is_boolean_type()")

        t = parse_type('[boolean]', self.schema)
        self.test_val("str(t)", "[boolean]")
        self.test_bool("t.get_element_type().is_boolean_type()")

        t = parse_type('{boolean : (any,any)}', self.schema)
        self.test_val("str(t)", "{ boolean : (any, any) }")
        self.test_bool("t.get_key_type().is_boolean_type()")
        self.test_bool("t.get_value_type().get_element_types()[0]."
                       "is_any_type()")

    def check_instance (self):
        "parse instance types: 9"
        from foo import Foo, bar

        t = parse_type('foo.Foo', self.schema)
        self.test_val("t.klass_name", "foo.Foo")
        self.test_val("str(t)", "foo.Foo")
        self.test_bool("t.is_instance_type()")

        t = parse_type('foo.bar.FooBar', self.schema)
        self.test_val("t.klass_name", "foo.bar.FooBar")
        self.test_val("str(t)", "foo.bar.FooBar")

        self.test_exc("parse_type('bar.FooBar', self.schema)", ValueError)
        self.test_exc("parse_type('Foo', self.schema)", ValueError)

        t = parse_type('Plain', self.schema)
        self.test_val("t.klass_name", "Plain")
        self.test_val("str(t)", "Plain")

    def check_list (self):
        "parse simple list types: 18"
        t = parse_type('[int]', self.schema)
        self.test_bool("t.is_container_type()")
        self.test_val("t.element_type.type", IntType)
        self.test_val("str(t)", "[int]")

        self.test_exc("parse_type('[StringType]', self.schema)", ValueError)
        self.test_exc("parse_type('[list]', self.schema)", ValueError)
        self.test_exc("parse_type('[string', self.schema)", ValueError)

        t = parse_type('[foo.Foo]', self.schema)
        self.test_bool("t.is_container_type()")
        self.test_bool("t.element_type.is_instance_type()")
        self.test_val("str(t)", "[foo.Foo]")

        self.test_exc("parse_type('[foo.FooBar]', self.schema)", ValueError)
        t = parse_type('[Plain]', self.schema)
        self.test_val("str(t)", "[Plain]")
        self.test_val("t.get_element_type().get_class_name()", "Plain")

        # should allow empty list (tuple, dict) someday, but for now we don't
        self.test_exc("parse_type('[]', self.schema)", ValueError)
        self.test_exc("parse_type('[int, string]', self.schema)", ValueError)

        # but as long as [any] works, [] isn't really needed!
        t = parse_type('[any]', self.schema)
        self.test_val("str(t)", "[any]")
        self.test_bool("t.is_container_type()")
        self.test_bool("t.is_any_type()", want_true=0)
        self.test_bool("t.get_element_type().is_any_type()")

    def check_tuple (self):
        "parse simple tuple types: 21"
        t = parse_type('(string,)', self.schema)
        self.test_val("str(t)", "(string,)")
        self.test_bool("t.is_container_type()")
        self.test_val("t.container_type", "tuple")
        self.test_val("len(t.element_types)", 1)
        self.test_val("t.element_types[0].type", StringType)

        t = parse_type('(int,string)', self.schema)
        self.test_val("str(t)", "(int, string)")
        self.test_bool("t.is_extended()", want_true=0)
        self.test_val("len(t.element_types)", 2)
        self.test_val("t.element_types[0].type", IntType)
        self.test_val("t.element_types[1].type", StringType)

        t = parse_type('(int, string*)', self.schema)
        self.test_val("str(t)", "(int, string*)")
        self.test_bool("t.is_extended()")
        self.test_val("len(t.element_types)", 2)
        self.test_val("t.element_types[0].type", IntType)
        self.test_val("t.element_types[1].type", StringType)

        self.test_exc("parse_type('(IntType)', self.schema)", ValueError)
        self.test_exc("parse_type('(IntType,)', self.schema)", ValueError)
        self.test_exc("parse_type('(list)', self.schema)", ValueError)

        self.test_exc("parse_type('()', self.schema)", ValueError)
        t = parse_type('(any, string)', self.schema)
        self.test_val("str(t)", "(any, string)")
        self.test_bool("t.get_element_types()[0].is_any_type()")

    def check_dict (self):
        "parse simple dictionary types: 12"
        t = parse_type('{float: string}', self.schema)
        self.test_val("str(t)", "{ float : string }")
        self.test_bool("t.is_container_type()")
        self.test_val("t.container_type", "dictionary")
        self.test_val("t.key_type.type", FloatType)
        self.test_val("t.value_type.type", StringType)

        self.test_exc("parse_type('{float:}', self.schema)", ValueError)
        self.test_exc("parse_type('{string: float ', self.schema)", ValueError)
        self.test_exc("parse_type('{IntType : string}', self.schema)", ValueError)

        self.test_exc("parse_type('{}', self.schema)", ValueError)
        self.test_exc("parse_type('{int:string, int:float}', self.schema)", ValueError)

        t = parse_type('{string : any}', self.schema)
        self.test_val("str(t)", "{ string : any }")
        self.test_bool("t.get_value_type().is_any_type()")

    def check_instcont (self):
        "parse instance-container types: 15"
        from foo import Foo
        t = parse_type('foo.Foo [int]', self.schema)
        self.test_bool("t.is_instance_type()")
        self.test_bool("t.is_container_type()")
        self.test_bool("t.is_instance_container_type()")
        self.test_val("str(t)", "foo.Foo [int]")
        self.test_val("t.get_class_name()", "foo.Foo")
        self.test_val("t.get_element_type().type", IntType)

        t = parse_type('foo.Foo (float,string)', self.schema)
        self.test_val("str(t)", "foo.Foo (float, string)")
        self.test_val("t.get_class_name()", "foo.Foo")
        self.test_val("[t.type for t in t.get_element_types()]",
                      [FloatType, StringType])

        t = parse_type('foo.Foo {string:int}', self.schema)
        self.test_val("str(t)", "foo.Foo { string : int }")
        self.test_val("t.get_class_name()", "foo.Foo")
        self.test_val("t.get_item_types()[0].get_type()", StringType)
        self.test_val("t.get_item_types()[1].get_type()", IntType)

        # An instance-container that contains a container
        t = parse_type('foo.Foo {string : (int,string)}', self.schema)
        self.test_val("str(t)", "foo.Foo { string : (int, string) }")
        # And one that contains another instance-container
        t = parse_type('foo.Foo { int : foo.bar.FooBar [string] }',
                       self.schema)
        self.test_val("str(t)", "foo.Foo { int : foo.bar.FooBar [string] }")

    def check_recursive_1 (self):
        "parse simple recursive type specifications: 11"
        t = parse_type('[ (int,float) ]', self.schema)
        self.test_val("str(t)", "[(int, float)]")
        self.test_bool("t.is_container_type()")
        subtype = t.element_type
        self.test_bool("subtype.is_container_type()")
        self.test_bool("(subtype.element_types[0].type,"
                       " subtype.element_types[1].type)",
                       (IntType, FloatType))

        t = parse_type('int | [string]', self.schema)
        self.test_val("str(t)", "int | [string]")
        self.test_bool("t.is_union_type()")
        self.test_val("len(t.union_types)", 2)
        (alt1, alt2) = t.union_types
        self.test_val("alt1.type", IntType)
        self.test_bool("alt2.is_container_type()")
        self.test_val("alt2.container_type", "list")
        self.test_val("alt2.element_type.type", StringType)

    def check_recursive_2 (self):
        "parse complex recursive type specifications: 8"

        # Of course, anyone who would use a dictionary mapping
        # (int,string) tuples to either foo.Foo or foo.bar.FooBar
        # instances in production code should be shot... ;-)

        t1 = parse_type("{(int,string):foo.Foo|foo.bar.FooBar}", self.schema)
        self.test_val("str(t1)",
                      "{ (int, string) : foo.Foo | foo.bar.FooBar }")
        (keytype, valuetype) = t1.get_item_types()
        self.test_val("str(keytype)", "(int, string)")
        self.test_val("str(valuetype)", "foo.Foo | foo.bar.FooBar")

        # Now this is just silly.
        t2 = parse_type("[{float|long : [(string, foo.Foo)]}]", self.schema)
        self.test_val("str(t2)",
                      "[{ float | long : [(string, foo.Foo)] }]")
        eltype = t2.element_type
        (keytype, valuetype) = eltype.get_item_types()
        self.test_val("str(keytype)", "float | long")
        self.test_val("str(valuetype)", "[(string, foo.Foo)]")
        
        # And this is plain gratuitous.
        t = UnionType(self.schema)
        t.set_union_types([t1, t2])
        self.test_val("str(t)", "%s | %s" % (t1, t2))

        # Not much better
        t = self.schema.make_tuple_type((t1, parse_type("int", self.schema)))
        self.test_val("str(t)", "(%s, int)" % t1)

    def check_errors (self):
        "parse erroneous names and dotted-named in various contexts: 12"
        # Mainly for checking the error messages and ensuring that we
        # hit every branch of 'parse_type()' and 'create_subtype()'
        msg_fmt = ("invalid type '%s': "
                   "not an atomic type, alias, or known class")
        exc = self.test_exc("parse_type('FooBar', self.schema)",
                            ValueError)
        self.test_val("str(exc)", msg_fmt % "FooBar")

        exc = self.test_exc("parse_type('foo.bar', self.schema)",
                            ValueError)
        self.test_val("str(exc)", msg_fmt % "foo.bar")

        exc = self.test_exc("parse_type('[FooBar]', self.schema)",
                            ValueError)
        self.test_val("str(exc)", msg_fmt % "FooBar")

        exc = self.test_exc("parse_type('FooBar|int', self.schema)",
                            ValueError)
        self.test_val("str(exc)", msg_fmt % "FooBar")

        exc = self.test_exc("parse_type('int|FooBar', self.schema)",
                            ValueError)
        self.test_val("str(exc)", msg_fmt % "FooBar")

        exc = self.test_exc("parse_type('[ding.dong]', self.schema)",
                            ValueError)
        self.test_val("str(exc)", msg_fmt % "ding.dong")

    def check_alias (self):
        "parse types with aliases: 11"
        schema = self.schema
        schema.add_alias("real", "int|float|long")
        schema.add_alias("range", "(real,real)")

        t = parse_type('{string:range}', schema)
        self.test_val("str(t)", "{ string : range }")
        #self.test_val("t.expand_aliases()",
        #              "{ string : (int | float | long, int | float | long) }")
        self.test_bool("t.is_alias_type()", want_true=0)

        vtype = t.get_value_type()
        self.test_bool("vtype.is_alias_type()")
        self.test_val("str(vtype)", "range")
        self.test_val("vtype.get_alias_type().container_type", "tuple")
        etypes = vtype.get_alias_type().get_element_types()
        self.test_bool("etypes[0].is_alias_type()")
        self.test_val("etypes[0].get_alias_type()", schema.get_alias("real"))
        self.test_bool("etypes[0].get_alias_type().is_union_type()")

        from foo import Foo
        schema.add_alias("Foo", "foo.Foo")
        t = parse_type('Foo [int]', schema)
        self.test_bool("t.is_instance_container_type()")
        self.test_val("t.get_class_name()", "foo.Foo")
        self.test_val("t.get_element_type().get_type()", IntType)


if __name__ == "__main__":
    (scenarios, options) = parse_args()
    run_scenarios (scenarios, options)






#!/usr/bin/env python

#
# Test script for the ValueType class family (AtomicType, InstanceType,
# etc.).  Does not test parsing of type specifications -- that's
# in test_parsetype.py.
#

__revision__ = "$Id: test_valuetype.py,v 1.49 2001/08/21 22:20:28 gward Exp $"

import string
from types import IntType, LongType, FloatType, StringType, ListType
from mems.tools.unittest import TestScenario, parse_args, run_scenarios
from grouch.schema import ClassDefinition, ObjectSchema
from grouch.valuetype import \
     AtomicType, UnionType, \
     InstanceType, ContainerType, InstanceContainerType, \
     AnyType, BooleanType, AliasType
from grouch.context import TypecheckContext

tested_modules = [ "grouch.valuetype" ]

# Set up a fake module hierarchy:
#   module 'foo' provides class Foo (and sub-module bar)
#   module 'foo.bar' provides class FooBar
import fake_module


class ValueTypeTest (TestScenario):

    def setup (self):
        self.schema = fake_module.make_schema()

    def shutdown (self):
        del self.schema


    test_cases = string.split("atomic instance "
                              "list tuple dict "
                              "union instcont misc alias "
                              "pickle"
                             )

    def check_atomic (self):
        "atomic types: 17"
        t = AtomicType(self.schema)
        self.test_stmt("t.set_type('int')")
        self.test_val("t.metatype", "atomic")
        self.test_bool("t.is_atomic_type()")
        self.test_val("t.type", IntType)
        self.test_val("t.get_type()", IntType)
        self.test_val("str(t)", "int")
        self.test_stmt("t.set_type('float')")
        self.test_val("str(t)", "float")
        self.test_val("t.get_metatype()", "atomic")

        t = AtomicType(self.schema)
        self.test_stmt("t.set_type(FloatType)")
        self.test_val("t.metatype", "atomic")
        self.test_val("t.type", FloatType)
        self.test_val("str(t)", "float")

        t = AtomicType(self.schema)
        self.test_exc("t.set_type('foo')", ValueError)
        self.test_exc("t.set_type('list')", ValueError)
        self.test_exc("t.set_type(ListType)", ValueError)
        self.test_val("str(t)", "*no type*")

    def check_instance (self):
        "instance types: 8"
        from foo import Foo
        from foo.bar import FooBar

        Foo_def = self.schema.get_class_definition(Foo)
        FooBar_def = self.schema.get_class_definition(FooBar)
        t = InstanceType(self.schema)
        self.test_stmt("t.set_class_name('foo.Foo')")
        self.test_val("str(t)", "foo.Foo")
        self.test_bool("t.is_instance_type()")
        self.test_bool("t.is_plain_instance_type()")
        self.test_val("t.get_metatype()", "instance")

        t = InstanceType(self.schema)
        self.test_stmt("t.set_class_name('foo.bar.FooBar')")
        self.test_val("str(t)", "foo.bar.FooBar")

        t = InstanceType(self.schema)
        self.test_exc("t.set_class_name(FooBar)", TypeError)

    def check_list (self):
        "simple list types: 14"
        from grouch.valuetype import ListType
        t = ListType(self.schema)
        self.test_stmt("t.set_element_type(IntType)")
        self.test_val("str(t)", "[int]")
        self.test_bool("isinstance(t.element_type, AtomicType)")
        self.test_val("str(t.element_type)", "int")
        self.test_bool("t.is_container_type()")
        self.test_bool("t.is_plain_container_type()")
        self.test_val("t.get_element_type().type", IntType)

        t = ListType(self.schema)
        self.test_exc("t.set_element_type(53)", TypeError)
        self.test_stmt("t.set_element_type('int')")
        self.test_val("str(t)", "[int]")
        self.test_val("t.element_type.type", IntType)

        t = ListType(self.schema)
        inttype = self.schema.make_atomic_type("int")
        self.test_stmt("t.set_element_type(inttype)")
        self.test_val("t.element_type", inttype)
        self.test_val("str(t)", "[int]")
        
    def check_tuple (self):
        "simple tuple types: 21"
        from grouch.valuetype import TupleType
        t = TupleType(self.schema)
        self.test_stmt("t.set_element_types((IntType,))")
        self.test_val("str(t)", "(int,)")
        self.test_val("t.element_types[0].type", IntType)
        self.test_val("t.get_element_types()[0].type", IntType)
        self.test_val("t.get_metatype()", "container")

        t = TupleType(self.schema)
        self.test_stmt("t.set_element_types(('int','string'))")
        self.test_val("str(t)", "(int, string)")
        self.test_val("t.element_types[0].type", IntType)
        self.test_val("t.element_types[1].type", StringType)
        
        t = TupleType(self.schema)
        self.test_stmt("t.set_element_types(('int',StringType))")
        self.test_val("str(t)", "(int, string)")
        self.test_val("t.element_types[0].type", IntType)
        self.test_val("t.element_types[1].type", StringType)
        self.test_bool("t.is_extended()", want_true=0)
        t.set_extended(1)
        self.test_bool("t.is_extended()")
        self.test_val("str(t)", "(int, string*)")

        t = TupleType(self.schema)
        self.test_exc("t.set_element_types(('list',))", ValueError)
        self.test_exc("t.set_element_types((42,))", TypeError)
        self.test_exc("t.set_element_types('int')", TypeError)
        self.test_exc("t.set_element_types(42)", TypeError)
        self.test_stmt("t.set_element_types(('int',))")

    def check_dict (self):
        "simple dictionary types: 15"
        from grouch.valuetype import DictionaryType
        t = DictionaryType(self.schema)
        self.test_stmt("t.set_item_types('int', FloatType)")
        self.test_val("t.key_type.type", IntType)
        self.test_val("t.value_type.type", FloatType)
        self.test_val("str(t)", "{ int : float }")
        self.test_val("t.get_key_type().type", IntType)
        self.test_val("t.get_value_type().type", FloatType)
        (keytype, valuetype) = t.get_item_types()
        self.test_val("keytype.get_type()", IntType)
        self.test_val("valuetype.get_type()", FloatType)

        t = DictionaryType(self.schema)
        self.test_exc("t.set_item_types('foo', 'string')", ValueError)
        self.test_exc("t.set_item_types(('int', 'string'))", TypeError)
        self.test_exc("t.set_item_types('list', 'string')", ValueError)
        self.test_exc("t.set_item_types('int', 42)", TypeError)
        self.test_stmt("t.set_item_types(StringType, 'int')")
        self.test_val("str(t)", "{ string : int }")

        t = DictionaryType(self.schema)
        self.test_exc("t.set_item_types(IntType, 'foo')",
                      ValueError)

    def check_union (self):
        "simple union types: 9"
        t = UnionType(self.schema)
        self.test_stmt("t.set_union_types([IntType, FloatType, LongType])")
        self.test_bool("t.is_union_type()")
        self.test_val("len(t.union_types)", 3)
        self.test_val("(t.union_types[0].type,"
                      " t.union_types[1].type,"
                      " t.union_types[2].type)",
                      (IntType, FloatType, LongType))
        self.test_val("len(t.get_union_types())", 3)
        self.test_val("str(t)", "int | float | long")

        t = UnionType(self.schema)
        self.test_exc("t.set_union_types(IntType)", TypeError)
        self.test_exc("t.set_union_types((IntType,))", TypeError)
        self.test_exc("t.set_union_types(['foo'])", ValueError)

    def check_instcont (self):
        "simple instance-container types: 20"
        t = InstanceContainerType(self.schema)
        import foo
        Foo_def = self.schema.get_class_definition("foo.Foo")

        self.test_stmt("t.set_class_name('foo.Foo')")
        ctype = self.schema.make_list_type(IntType)
        self.test_stmt("t.set_container_type(ctype)")
        self.test_bool("t.is_instance_type()")
        self.test_bool("t.is_container_type()")
        self.test_bool("t.is_instance_container_type()")
        self.test_bool("t.is_plain_instance_type()", want_true=0)
        self.test_bool("t.is_plain_container_type()", want_true=0)
        self.test_val("t.get_metatype()", "instance-container")

        self.test_val("str(t)", "foo.Foo [int]")
        self.test_val("t.get_element_type().type", IntType)
        self.test_exc("t.get_element_types()", RuntimeError)
        self.test_exc("t.get_item_types()", RuntimeError)

        t = InstanceContainerType(self.schema)
        self.test_stmt("t.set_class_name('foo.Foo')")
        ctype = self.schema.make_tuple_type((IntType, FloatType))
        self.test_stmt("t.set_container_type(ctype)")
        self.test_val("str(t)", "foo.Foo (int, float)")
        etypes = t.get_element_types()
        self.test_val("(etypes[0].type, etypes[1].type)",
                      (IntType, FloatType))

        t = InstanceContainerType(self.schema)
        foo_type = InstanceType(self.schema)
        foo_type.set_class_name('foo.Foo')
        self.test_stmt("t.set_class_name('foo.Foo')")
        ctype = self.schema.make_dictionary_type(StringType, foo_type)
        self.test_stmt("t.set_container_type(ctype)")
        self.test_val("str(t)", "foo.Foo { string : foo.Foo }")
        (key_type, value_type) = t.get_item_types()
        self.test_val("key_type.get_type()", StringType)

    def check_misc (self):
        "miscellaneous type (any, boolean): 5"
        t = AnyType(self.schema)
        self.test_bool("t.is_any_type()")
        self.test_val("str(t)", "any")

        t = BooleanType(self.schema)
        self.test_bool("t.is_boolean_type()")
        self.test_bool("t.is_any_type()", want_true=0)
        self.test_val("str(t)", "boolean")
        
    def check_alias (self):
        "type aliasing: 4"
        schema = self.schema
        schema.add_alias("real", "int|float|long")

        t = schema.parse_type("real")
        self.test_bool("t.is_alias_type()")
        alias = t.get_alias_type()
        self.test_bool("alias.is_union_type()")
        self.test_val("str(t)", "real")

        types = alias.get_union_types()
        self.test_val("map(str, types)", ["int", "float", "long"])

    def check_pickle (self):
        "pickling/unpickling of ValueType objects: 20"
        from pickle import dumps, loads
        
        schema = self.schema
        parse_type = schema.parse_type
        t1 = parse_type("int")
        t2 = loads(dumps(t1)) ; t2._fixup_atomic_types({})

        # Let's make sure we understand the pre-pickle situation.
        self.test_bool("t1.schema is schema")
        self.test_bool("schema.get_type(IntType).schema is schema")

        self.test_val("str(t2)", "int")

        # Make sure that the easily-compared innards of the two schemata
        # match.  (Some attributes are full of non-comparable objects, like
        # klass_definitions -- a dict of ClassDefinitions.  We just skip
        # them out of laziness.)
        sch1 = t1.schema ; sch2 = t2.schema
        self.test_bool("sch1.atomic_type_names == sch2.atomic_type_names")
        self.test_bool("sch1.atomic_type_objects == sch2.atomic_type_objects")
        self.test_bool("sch1.atomic_type_values == sch2.atomic_type_values")

        self.test_bool("sch2.get_type(IntType).schema is sch2")
        self.test_bool("sch2.get_type(FloatType).schema is sch2")

        # The schemata are not directly comparable, so remove them from
        # consideration (now that I've verified that most of their innards
        # match).
        t1s = t1.schema ; t2s = t2.schema
        del t1.schema, t2.schema
        self.test_bool("t1.__dict__ == t2.__dict__")
        t1.schema = t1s ; t2.schema = t2s

        t1 = parse_type("[string | float]")
        t2 = loads(dumps(t1)) ; t2._fixup_atomic_types({})
        self.test_val("str(t2)", "[string | float]")

        from mx.DateTime import now
        schema = ObjectSchema()
        schema.add_alias("real", "int|long|float")
        schema.add_atomic_type(now())
        t1 = schema.parse_type("real")
        t2 = schema.parse_type("(string, DateTime)")
        
        t1b = loads(dumps(t1)) ; t1b._fixup_atomic_types({})
        t2b = loads(dumps(t2)) ; t2b._fixup_atomic_types({})

        self.test_val("str(t1b)", "real")
        self.test_bool("t1b.is_alias_type()")
        self.test_bool("t1b.get_alias_type().is_union_type()")

        self.test_val("str(t2b)", "(string, DateTime)")
        self.test_bool("t2b.is_container_type()")
        self.test_bool("t2b.get_element_types()[1].is_atomic_type()")

        schema1 = schema
        schema2 = loads(dumps(schema1))
        innertype1 = schema1.alias_value['real'].union_types[0]
        innertype2 = schema2.alias_value['real'].union_types[0]
        self.test_bool("schema1.alias_value['real'].schema is schema1")
        self.test_bool("schema2.alias_value['real'].schema is schema2")
        self.test_bool("innertype1.schema is schema1")
        self.test_bool("innertype2.schema is schema2")

# class ValueTypeTest


class TypecheckTest (TestScenario):
    
    def setup (self):
        self.context = TypecheckContext()
        self.context.report_errors = 0

        from foo import Foo
        from foo.bar import FooBar
        from fake_module import make_schema
        
        self.schema = make_schema()
        self.foo_type = self.schema.make_instance_type(Foo)
        self.foobar_type = self.schema.make_instance_type(FooBar)

        # Create canonical instances with known good values.
        self.foo = Foo()
        self.foobar = FooBar()

    def shutdown (self):
        del self.context, self.schema
        del self.foo_type, self.foo


    test_cases = string.split("atomic instance_1 instance_2 instance_3 "
                              "list tuple dict "
                              "constraints instcont union "
                              "any boolean alias "
                              "nested_1 nested_2 recursive")

    def check_atomic (self):
        "type-checking atomic values: 13"

        t1 = self.schema.parse_type("int")
        t2 = self.schema.parse_type("string")
        ctx = self.context

        self.test_bool("t1.check_value(3, ctx)")
        self.test_bool("t1.check_value(3.0, ctx)", want_true=0)
        self.test_val("ctx.num_errors()", 1)
        self.test_val("ctx.errors",
                      [("", "expected int, got float (3.0)")])
        self.test_bool("t1.check_value(None, ctx)")
        self.test_val("ctx.visit_counts", {'atomic': {'int': 3}})

        self.test_bool("t2.check_value('', ctx)")
        self.test_bool("t2.check_value(3, ctx)", want_true=0)
        self.test_val("ctx.num_errors()", 2)
        self.test_val("ctx.errors",
                      [("", "expected int, got float (3.0)"),
                       ("", "expected string, got int (3)")])
        self.test_val("ctx.format_errors()",
                      ["expected int, got float (3.0)",
                       "expected string, got int (3)"])
        self.test_val("ctx.visit_counts", {'atomic': {'int': 3, 'string': 2}})

        self.test_bool("t1.check_value(None, ctx)")

        #t1.allow_none = 0
        #self.test_bool("t1.check_value(None, ctx)", want_true=0)

    def check_instance_1 (self):
        "type-checking of simple instance values: 18"

        ctx = self.context ; foo = self.foo ; foo_type = self.foo_type

        # No errors: default values are OK
        self.test_bool("foo_type.check_value(foo, ctx)")
        self.test_val("ctx.num_errors()", 0)
        self.test_val("ctx.visit_counts",
                      {'atomic': {'int': 1, 'string': 1},
                       'instance': {'foo.Foo': 1}})

        # One bad attribute (add to empty error list)
        foo.num = 3.7
        msg = "expected int, got float (%s)" % repr(3.7)
        ctx.reset()
        self.test_bool("foo_type.check_value(foo, ctx)", want_true=0)
        self.test_val("ctx.errors", [("num", msg)])
        self.test_val("ctx.format_errors()", ["num: " + msg])

        # Two bad attributes
        ctx.reset()
        foo.num = "boo"
        foo.msg = 0
        self.test_bool("foo_type.check_value(foo, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("num", "expected int, got string ('boo')"),
                       ("msg", "expected string, got int (0)")])
        self.test_val("ctx.format_errors()",
                      ["num: expected int, got string ('boo')",
                       "msg: expected string, got int (0)"])
        self.test_val("ctx.visit_counts",
                      {'atomic': {'int': 1, 'string': 1},
                       'instance': {'foo.Foo': 1}})

        # One definitely good, one None (should give no errors)
        ctx.reset()
        foo.num = 666
        foo.msg = None
        self.test_bool("foo_type.check_value(foo, ctx)")
        self.test_val("ctx.num_errors()", 0)

        # The "instance" itself is None
        self.test_bool("foo_type.check_value(None, ctx)")
        self.test_val("ctx.num_errors()", 0)

        # And the "instance" is not an instance or None
        self.test_bool("foo_type.check_value(42, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("", "expected instance of foo.Foo, got int (42)")])

        class Blah: pass
        ctx.reset()
        blah = Blah()
        self.test_bool("foo_type.check_value(blah, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("",
                        "expected instance of foo.Foo, got instance of Blah"),
                       ("",
                        "found instance of class not in schema: Blah")])


    def check_instance_2 (self):
        "type-checking of instance values with missing/unexpected attrs: 22"

        ctx = self.context ; foo = self.foo ; foo_type = self.foo_type

        # One attribute missing
        del foo.msg
        self.test_bool("foo_type.check_value(foo, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("", "expected attribute 'msg' not found")])
        self.test_val("ctx.visit_counts",
                      {'atomic': {'int': 1},
                       'instance': {'foo.Foo': 1}})
        
        # One attribute missing, one with bad type
        ctx.reset()
        del foo.num
        foo.msg = 0
        self.test_bool("foo_type.check_value(foo, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("", "expected attribute 'num' not found"),
                       ("msg", "expected string, got int (0)")])
        self.test_val("ctx.visit_counts",
                      {'atomic': {'string': 1},
                       'instance': {'foo.Foo': 1}})

        # One unexpected attribute
        ctx.reset()
        foo.num = 1
        foo.msg = "Foo!"
        foo.c = 6.02e23
        unexp_msg = "found unexpected attribute 'c': float (%s)" % `foo.c`
        self.test_bool("foo_type.check_value(foo, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("", unexp_msg)])
        self.test_val("ctx.visit_counts",
                      {'atomic': {'int': 1, 'string': 1},
                       'instance': {'foo.Foo': 1}})

        # One unexpected attribute, one bad value
        ctx.reset()
        foo.num = 1L
        self.test_bool("foo_type.check_value(foo, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("", unexp_msg),
                       ("num", "expected int, got long int (1L)")])
        del foo.c

        # Mixups between instance and class attributes.
        from foo import Foo
        ctx.reset()
        del foo.num
        Foo.num = 37
        self.test_val("foo.num", 37)    # normal Python semantics are OK
        self.test_bool("getattr(foo, 'num')", 37)
        self.test_bool("hasattr(foo, 'num')")
        self.test_bool("foo_type.check_value(foo, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("", "expected attribute 'num' not found")])

        ctx.reset()
        foo.num = 42
        self.test_val("foo.num", 42)
        self.test_bool("foo_type.check_value(foo, ctx)")
        Foo.x = "boo"
        self.test_bool("foo_type.check_value(foo, ctx)")
        del Foo.num
        self.test_bool("foo_type.check_value(foo, ctx)")
        del Foo.x
        self.test_bool("foo_type.check_value(foo, ctx)")
        self.test_val("ctx.errors", [])

    def check_instance_3 (self):
        "type-checking of instances of subclasses: 12"
        from foo import Foo, SubFoo
        ctx = self.context
        foo_type = self.foo_type
        foo = Foo()
        subfoo = SubFoo()

        self.test_bool("foo_type.check_value(foo, ctx)")
        self.test_bool("foo_type.check_value(subfoo, ctx)")
        self.test_val("ctx.num_errors()", 0)
        self.test_val("ctx.visit_counts",
                      {'atomic': {'int': 3, 'string': 2},
                       'instance': {'foo.Foo': 2}})

        subfoo.num = "boo"
        ctx.reset(start_name='subfoo')
        self.test_bool("foo_type.check_value(subfoo, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("subfoo.num", "expected int, got string ('boo')")])

        ctx.reset(start_name='subfoo')
        subfoo.num = 0
        subfoo.k = "hi"
        subfoo.i = None
        self.test_bool("foo_type.check_value(subfoo, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("subfoo", "found unexpected attribute 'i': None"),
                       ("subfoo.k", "expected int, got string ('hi')")])

        ctx.reset(start_name='subfoo')
        subfoo = SubFoo()
        del subfoo.k, subfoo.num
        self.test_bool("foo_type.check_value(subfoo, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("subfoo", "expected attribute 'num' not found"),
                       ("subfoo", "expected attribute 'k' not found")])

        # Cook up a sub-class that's not known to the schema (to test
        # one measly line of code!)
        from fake_module import move_class
        class AltFoo (Foo): pass
        move_class("foo", AltFoo)      # now it's foo.AltFoo
        AltFoo_def = ClassDefinition("foo.AltFoo",
                                      self.schema,
                                      bases=["foo.Foo"])
        AltFoo_def.finish_definition()

        ctx.reset(start_name='alt_foo')
        alt_foo = AltFoo()
        self.test_bool("foo_type.check_value(alt_foo, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("alt_foo",
                        "expected instance of foo.Foo, "
                        "got instance of unknown subclass foo.AltFoo")])

    def check_list (self):
        "type-checking of list values: 12"
        ctx = self.context
        list_type = self.schema.make_list_type('int')

        self.test_bool("list_type.check_value([], ctx)")
        self.test_val("ctx.num_errors()", 0)
        self.test_bool("list_type.check_value([0, -34], ctx)")
        self.test_val("ctx.num_errors()", 0)
        self.test_val("ctx.visit_counts",
                      {'container': {'list': 2},
                       'atomic': {'int': 2}})

        ctx.reset("L")
        self.test_bool("list_type.check_value([0, 3.4], ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("L[1]", "expected int, got float (%s)" % repr(3.4))])
        self.test_val("ctx.visit_counts",
                      {'container': {'list': 1},
                       'atomic': {'int': 2}})

        ctx.reset("L")
        self.test_bool("list_type.check_value(['foo', 0L], ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("L[0]", "expected int, got string ('foo')"),
                       ("L[1]", "expected int, got long int (0L)")])

        ctx.reset("L")
        self.test_bool("list_type.check_value((3, 4), ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("L", "expected list, got tuple ((3, 4))")])

    def check_tuple (self):
        "type-checking of tuple values: 21"
        ctx = self.context
        tuple_type = self.schema.make_tuple_type(('string', 'int'))

        self.test_bool("tuple_type.check_value(('foo', 3), ctx)")
        self.test_val("ctx.num_errors()", 0)
        self.test_val("ctx.visit_counts",
                      {'container': {'tuple': 1},
                       'atomic': {'string': 1, 'int': 1}})

        ctx.reset()
        self.test_bool("tuple_type.check_value(['foo', 3], ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("", "expected tuple, got list (['foo', 3])")])
        self.test_val("ctx.visit_counts",
                      {'container': {'tuple': 1}})

        ctx.reset()
        self.test_bool("tuple_type.check_value((), ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("","expected tuple of length 2, got one of length 0")])

        ctx.reset()
        self.test_bool("tuple_type.check_value(('foo', 3L), ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("[1]", "expected int, got long int (3L)")])

        ctx.reset()
        self.test_bool("tuple_type.check_value((3, 'foo'), ctx)",want_true=0)
        self.test_val("ctx.errors",
                      [("[0]", "expected string, got int (3)"),
                       ("[1]", "expected int, got string ('foo')")])

        ctx.reset()
        tuple_type.set_extended(1)
        self.test_bool("tuple_type.check_value(('foo', 3), ctx)")
        self.test_bool("tuple_type.check_value(('foo',), ctx)")
        self.test_bool("tuple_type.check_value(('foo', 3, 4, 5), ctx)")
        self.test_bool("tuple_type.check_value(('foo',''), ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("[1]", "expected int, got string ('')")])

        ctx.reset()
        self.test_bool("tuple_type.check_value((), ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("",
                        "expected tuple of at least length 1, "
                        "got one of length 0")])

        ctx.reset()
        self.test_bool("tuple_type.check_value(('foo',3,666,''), ctx)",
                       want_true=0)
        self.test_val("ctx.errors",
                      [("[3]", "expected int, got string ('')")])

    def check_dict (self):
        "type-checking of dictionary values: 16"
        ctx = self.context
        dict_type = self.schema.make_dictionary_type('string', 'long')

        self.test_bool("dict_type.check_value({'foo': 3L, 'bar': 666L}, ctx)")
        self.test_val("ctx.num_errors()", 0)
        self.test_bool("dict_type.check_value({}, ctx)")
        self.test_val("ctx.num_errors()", 0)
        self.test_val("ctx.visit_counts",
                      {'container': {'dictionary': 2},
                       'atomic': {'string': 2, 'long': 2}})
                
        # Bad value types
        ctx.reset()
        self.test_bool("dict_type.check_value({'foo': 3}, ctx)",
                       want_true=0)
        self.test_val("ctx.errors",
                      [("['foo']", "expected long, got int (3)")])
        self.test_val("ctx.visit_counts",
                      {'container': {'dictionary': 1},
                       'atomic': {'string': 1, 'long': 1}})

        ctx.reset()
        self.test_bool("dict_type.check_value({'foo': 3L, 'bar': 3.14}, ctx)",
                       want_true=0)
        self.test_val("ctx.errors",
                      [("['bar']",
                        "expected long, got float (%s)" % repr(3.14))])

        # Bad key types
        ctx.reset()
        self.test_bool("dict_type.check_value({1: 3L}, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("[1] (key)", "expected string, got int (1)")])

        # A bit of each
        ctx.reset()
        self.test_bool("dict_type.check_value({1: 'foo'}, ctx)", want_true=0)
        self.test_seq("ctx.errors",
                      [("[1] (key)", "expected string, got int (1)"),
                       ("[1]", "expected long, got string ('foo')")],
                      match_order=0)

        # Same, but in different elements
        ctx.reset()
        self.test_bool("dict_type.check_value({1: 1L, 'foo': 'foo'}, ctx)",
                       want_true=0)
        self.test_seq("ctx.errors",
                      [("[1] (key)", "expected string, got int (1)"),
                       ("['foo']", "expected long, got string ('foo')")
                      ], match_order=0)
        
    def check_constraints (self):
        "type-checking with constraints (not none, not empty): 19"
        ctx = self.context
        int_type = self.schema.parse_type("int")
        int_type.allow_none = 1
        self.test_bool("int_type.check_value(None, ctx)")
        self.test_val("ctx.num_errors()", 0)

        ctx.reset()
        int_type.allow_none = 0
        self.test_bool("int_type.check_value(None, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("", "expected int, got None")])

        ctx.reset()
        list_type = self.schema.parse_type("[int]")
        self.test_bool("list_type.check_value(None, ctx)")
        self.test_val("ctx.num_errors()", 0)
        self.test_bool("list_type.check_value([], ctx)")
        self.test_val("ctx.num_errors()", 0)

        ctx.reset()
        list_type.allow_none = 0
        self.test_bool("list_type.check_value(None, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("", "expected list, got None")])
        ctx.reset()
        self.test_bool("list_type.check_value([], ctx)")
        self.test_val("ctx.num_errors()", 0)

        ctx.reset()
        list_type.allow_empty = 0
        self.test_bool("list_type.check_value([], ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("", "empty list not allowed")])
        ctx.reset()
        self.test_bool("list_type.check_value(None, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("", "expected list, got None")])

        ctx.reset()
        list_type.allow_none = 1
        self.test_bool("list_type.check_value([], ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("", "empty list not allowed")])
        ctx.reset()
        self.test_bool("list_type.check_value(None, ctx)")
        
    def check_instcont (self):
        "type-checking instance-container types: 11"
        import UserList
        ctx = self.context
        ulist_def = ClassDefinition("UserList.UserList", self.schema)
        ulist_def.add_attribute("data", self.schema.parse_type("[any]"))
        ulist_def.finish_definition()
        self.schema.add_class(ulist_def)
        ulist_type = self.schema.parse_type("UserList.UserList [string]")

        good_list = UserList.UserList(["foo", "bar"])
        empty_list = UserList.UserList()
        bad_list = UserList.UserList([3, "foo"])
        
        self.test_bool("ulist_type.check_value(good_list, ctx)")
        self.test_val("ctx.num_errors()", 0)
        self.test_bool("ulist_type.check_value(empty_list, ctx)")
        self.test_val("ctx.num_errors()", 0)
        self.test_bool("ulist_type.check_value(bad_list, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("[0]", "expected string, got int (3)")])
        self.test_val("ctx.visit_counts",
                      {'instance-container': {'UserList.UserList': 3},
                       'container': {'list': 3},
                       'atomic': {'string': 4},
                       'any': {None: 4}})

        ctx.reset()
        self.test_bool("ulist_type.check_value(['foo'], ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("", "expected instance of UserList.UserList, "
                            "got list (['foo'])")])
        
        ctx.reset()
        self.test_bool("ulist_type.check_value('foo', ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("", "expected instance of UserList.UserList, "
                            "got string ('foo')")])
        
    def check_union (self):
        "type-checking of union types: 28"
        ctx = self.context
        union_type = UnionType(self.schema)

        # Start with a degenerate union type: the opposite of "any"
        # (ie. nothing is valid).
        union_type.set_union_types([])
        self.test_bool("union_type.check_value(0, ctx)", want_true=0)
        self.test_bool("union_type.check_value('hi', ctx)", want_true=0)
        self.test_val("ctx.num_errors()", 2)
        self.test_val("ctx.errors[0][1]",
                      "no valid values for empty union type; got int (0)")
        self.test_val("ctx.errors[1][1]",
                      "no valid values for empty union type; "
                      "got string ('hi')")
        self.test_val("ctx.visit_counts",
                      {'union': {'': 2}})

        ctx.reset()
        self.test_bool("union_type.check_value(ctx, ctx)", want_true=0)
        self.test_val("ctx.num_errors()", 2)
        self.test_val("ctx.errors[0][1]",
                      "no valid values for empty union type; "
                      "got instance of grouch.context.TypecheckContext")
        self.test_val("ctx.errors[1][1]",
                      "found instance of class not in schema: "
                      "grouch.context.TypecheckContext")

        # Equally pointless: just one type in the union.
        ctx.reset()
        union_type.set_union_types([self.schema.get_type('int')])
        self.test_bool("union_type.check_value(0, ctx)")
        self.test_val("ctx.num_errors()", 0)
        self.test_bool("union_type.check_value('hi', ctx)", want_true=0)
        self.test_val("ctx.num_errors()", 1)
        self.test_val("ctx.errors[0][1]", "expected int; got string ('hi')")
        self.test_val("ctx.visit_counts",
                      {'union': {'int': 2}})

        # Finally, something that would occur in reality: int|long
        ctx.reset()
        union_type.set_union_types([self.schema.get_type('int'),
                                    self.schema.get_type('long')])
        self.test_bool("union_type.check_value(100, ctx)")
        self.test_bool("union_type.check_value(2L**40, ctx)")
        self.test_val("ctx.num_errors()", 0)
        self.test_bool("union_type.check_value(1.0, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("", "expected int or long; got float (1.0)")])
        self.test_val("ctx.visit_counts",
                      {'union': {'int | long': 3}})

        # Now one with three types, just to make sure the error message
        # is formatted correctly.
        ctx.reset()
        union_type.set_union_types([self.schema.get_type('int'),
                                    self.schema.get_type('long'),
                                    self.schema.get_type('float')])
        self.test_bool("union_type.check_value(100, ctx)")
        self.test_bool("union_type.check_value(2L**40, ctx)")
        self.test_bool("union_type.check_value(1.0, ctx)")
        self.test_val("ctx.num_errors()", 0)
        self.test_bool("union_type.check_value('foo', ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("", "expected int, long, or float; "
                        "got string ('foo')")])

    # check_union ()

    def check_any (self):
        "type-checking of 'any' type: 19"
        ctx = self.context
        any_type = AnyType(self.schema)
        self.test_bool("any_type.check_value(3, ctx)")
        self.test_bool("any_type.check_value(3.0, ctx)") 
        self.test_bool("any_type.check_value(3L, ctx)")
        self.test_bool("any_type.check_value('three', ctx)")
        self.test_bool("any_type.check_value(self, ctx)")
        self.test_bool("any_type.check_value(self.__class__, ctx)")
        self.test_bool("any_type.check_value((self, self.check_any), ctx)")
        self.test_val("ctx.num_errors()", 0)
        self.test_val("ctx.visit_counts",
                      {'any': {None: 7}})

        # This won't work right now, 'cause AnyType accepts *anything*,
        # even if None isn't allowed.  Hmmm.
        ctx.reset()
        any_type.allow_none = 0
        self.test_bool("any_type.check_value(3, ctx)")
        self.test_bool("any_type.check_value(None, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("", "expected anything except None")])

        # Ensure that we typecheck instances when seen in an "any" slot.
        from foo import Foo
        ctx.reset()
        f = Foo() ; f.num = "37"
        self.test_bool("any_type.check_value(f, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("num", "expected int, got string ('37')")])
        self.test_val("ctx.visit_counts",
                      {'any': {None: 1},
                       'instance': {'foo.Foo': 1},
                       'atomic': {'int': 1, 'string': 1}})

        # Ensure that instances of "foreign" classes (classes not in
        # the schema) don't cause any problems by default.
        ctx.reset()
        self.test_bool("any_type.check_value(ctx, ctx)")
        self.test_val("ctx.num_errors()", 0)

        # But if we set the 'allow_any_instance' flag to false, foreign
        # instances are trouble.
        ctx.reset()
        any_type.set_allow_any_instance(0)
        self.test_bool("any_type.check_value(ctx, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("", "found instance of class not in schema: "
                            "grouch.context.TypecheckContext")])

    def check_boolean (self):
        "type-checking of boolean type: 11" # 13
        ctx = self.context
        bool_type = BooleanType(self.schema)
        self.test_bool("bool_type.check_value(0, ctx)")
        self.test_bool("bool_type.check_value(1, ctx)")
        self.test_bool("bool_type.check_value(None, ctx)")
        self.test_val("ctx.num_errors()", 0)
        self.test_val("ctx.visit_counts",
                      {'boolean': {None: 3}})

        self.test_bool("bool_type.check_value('', ctx)", want_true=0)
        self.test_bool("bool_type.check_value('foo', ctx)", want_true=0)
        self.test_bool("bool_type.check_value(2, ctx)", want_true=0)
        errs = ctx.errors
        self.test_val("errs[0]",
                      ("", "expected a boolean, got string ('')"))
        self.test_val("errs[1]",
                      ("", "expected a boolean, got string ('foo')"))
        self.test_val("errs[2]",
                      ("", "expected a boolean, got int (2)"))

        # Again, allow_none is currently ignored -- hmmm!
        #ctx.reset()
        #bool_type.allow_none = 0
        #self.test_bool("bool_type.check_value(None, ctx)", want_true=0)
        #self.test_val("ctx.errors",
        #              [("", "expected a boolean, got None")])

    def check_alias (self):
        "type-checking of alias types: 6"
        ctx = self.context
        schema = self.schema
        alias_type = AliasType(schema)
        alias_type.set_alias_type("real", schema.parse_type("int|long|float"))
        
        self.test_bool("alias_type.check_value(100, ctx)")
        self.test_bool("alias_type.check_value(2L**40, ctx)")
        self.test_bool("alias_type.check_value(3.14159, ctx)")
        self.test_val("ctx.num_errors()", 0)
        self.test_bool("alias_type.check_value('foo', ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("", "expected int, long, or float; "
                        "got string ('foo')")])

    def check_nested_1 (self):
        "type-checking of nested container structures: 10"

        ctx = self.context
        ntype = self.schema.parse_type("{string : [int]}")

        # val1 - completely good value
        # val2 - bad at top level (not a list)
        # val3 - bad deeper down (a list, but not of ints)
        # val4 - bad in both ways
        val1 = {"foo": [1, 2], "bar": [0]}
        val2 = {"foo": 1, "bar": [0]}
        val3 = {"foo": [4, 5], "bar": [1.0]}
        val4 = {"foo": "bar", "bar": [4, 2L]}

        ctx.start_context('val')
        self.test_bool("ntype.check_value(val1, ctx)")
        self.test_val("ctx.num_errors()", 0)

        self.test_bool("ntype.check_value(val2, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("val['foo']", "expected list, got int (1)")])

        ctx.reset('val')
        self.test_bool("ntype.check_value(val3, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("val['bar'][0]", "expected int, got float (1.0)")])

        ctx.reset('val')
        self.test_bool("ntype.check_value(val4, ctx)", want_true=0)
        ctx.errors.sort()
        self.test_val("ctx.num_errors()", 2)
        self.test_val("ctx.errors[0]",
                      ("val['bar'][1]", "expected int, got long int (2L)"))
        self.test_val("ctx.errors[1]",
                      ("val['foo']", "expected list, got string ('bar')"))

    def check_nested_2 (self):
        "type-checking of nested instances: 24"

        from __main__ import Plain
        from foo import SubFoo
        from foo.bar import FooBar
        ctx = self.context
        foobar_type = self.foobar_type

        
        # Schema for Foo and FooBar classes (see also fake_module.py):
        # class foo.Foo:
        #     num : int = 37
        #     msg : msg = "hello!"
        #
        # class foo.bar.FooBar:
        #     a : float = 3.14159
        #     b : Foo = None

        # Get initial objects with default values -- will only check
        # foobar because foobar.b is None.
        foo = self.foo ; foobar = self.foobar
        subfoo = SubFoo()
        ctx.start_context('foobar')
        self.test_bool("foobar_type.check_value(foobar, ctx)")
        self.test_val("ctx.num_errors()", 0)

        # Now link the two: still should not have any errors!
        foobar.b = foo
        self.test_bool("foobar_type.check_value(foobar, ctx)")
        self.test_val("ctx.num_errors()", 0)

        # Now make 'b' an instance of SubFoo
        foobar.b = subfoo
        self.test_bool("foobar_type.check_value(foobar, ctx)")
        self.test_val("ctx.num_errors()", 0)

        # Try a couple of different bad values for 'b'; all of these should
        # generate exactly one error, since they're not instances whose
        # class the schema knows anything about.
        foobar.b = 666
        ctx.reset(start_name='foobar')
        self.test_bool("foobar_type.check_value(foobar, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("foobar.b",
                        "expected instance of foo.Foo, got int (666)")])

        ctx.reset(start_name='foobar')
        foobar.b = [1, 'hello']
        self.test_bool("foobar_type.check_value(foobar, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("foobar.b",
                        "expected instance of foo.Foo, "
                        "got list ([1, 'hello'])")])

        ctx.reset(start_name='foobar')
        foobar.b = Plain()
        self.test_bool("foobar_type.check_value(foobar, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("foobar.b",
                        "expected instance of foo.Foo, got instance of Plain")
                      ])

        # Back to 'b' being (properly) a Foo instance, but now give it a
        # bad attribute value or two.
        ctx.reset()
        foobar.b = foo
        foo.num = 1.0
        self.test_bool("foobar_type.check_value(foobar, ctx)", want_true=0)
        self.test_val("ctx.format_errors()",
                      [("b.num: expected int, got float (1.0)")])

        ctx.reset()
        foo.num = FooBar()
        self.test_bool("foobar_type.check_value(foobar, ctx)", want_true=0)
        self.test_val("ctx.format_errors()",
                      [("b.num: expected int, "
                        "got instance of foo.bar.FooBar")])

        # Make sure that we check instances of known classes, even if
        # we stumble across them in the wrong place.
        ctx.reset()
        foo.num.b = 1
        self.test_bool("foobar_type.check_value(foobar, ctx)", want_true=0)
        self.test_val("ctx.num_errors()", 2)
        self.test_val("ctx.errors[0]",
                      ("b.num",
                       "expected int, got instance of foo.bar.FooBar"))
        self.test_val("ctx.errors[1]",
                      ("b.num.b",
                       "expected instance of foo.Foo, got int (1)"))

        # Make sure that we detect errors when foobar.b is an instance
        # of SubFoo rather than Foo directly.
        ctx.reset(start_name='foobar')
        foobar.b = subfoo
        foobar.b.num = "3"
        self.test_bool("foobar_type.check_value(foobar, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("foobar.b.num",
                        "expected int, got string ('3')")])
        foobar.b.k = "FOO"
        ctx.reset(start_name='foobar')
        self.test_bool("foobar_type.check_value(foobar, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("foobar.b.num",
                        "expected int, got string ('3')"),
                       ("foobar.b.k",
                        "expected int, got string ('FOO')")])

    def check_recursive (self):
        "type-check recursive data structures: 16"

        from foo import TreeNode
        ctx = self.context
        tn_type = self.schema.make_instance_type(TreeNode)

        # Yawn, boring, trivial.
        parent = TreeNode()
        self.test_bool("tn_type.check_value(parent, ctx)")
        self.test_val("ctx.num_errors()", 0)

        # Let's get a bit more interesting.
        child = TreeNode()
        parent.children.append(child)
        child.parent = parent
        self.test_bool("tn_type.check_value(parent, ctx)")
        self.test_val("ctx.num_errors()", 0)
        self.test_bool("tn_type.check_value(child, ctx)")
        self.test_val("ctx.num_errors()", 0)

        # Add some errors to make things more fun.
        parent.name = 666
        ctx.reset(start_name='parent')
        self.test_bool("tn_type.check_value(parent, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("parent.name", "expected string, got int (666)")])

        ctx.reset(start_name='parent')
        parent.name = "foo"
        child.name = 1
        self.test_bool("tn_type.check_value(parent, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("parent.children[0].name",
                        "expected string, got int (1)")])

        ctx.reset(start_name='parent')
        child.name = child
        self.test_bool("tn_type.check_value(parent, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("parent.children[0].name",
                        "expected string, got instance of foo.TreeNode")])

        ctx.reset(start_name='child')
        self.test_bool("tn_type.check_value(child, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("child.name",
                        "expected string, got instance of foo.TreeNode")])

        ctx.reset(start_name='child')
        child.name = "bar"
        parent.children.append(1)
        self.test_bool("tn_type.check_value(child, ctx)", want_true=0)
        self.test_val("ctx.errors",
                      [("child.parent.children[1]",
                        "expected instance of foo.TreeNode, got int (1)")])

    # check_recursive


# class TypecheckTest


if __name__ == "__main__":
    (scenarios, options) = parse_args()
    run_scenarios (scenarios, options)



#!/usr/bin/env python

# 
# Test script for the grouch.util module.
# 
# created 2001/03/09, GPW
#

# XXX currently requires ExtensionClass!

__revision__ = "$Id: test_util.py,v 1.2 2001/08/21 22:20:28 gward Exp $"


from ExtensionClass import ExtensionClass, Base
from mems.tools.unittest import TestScenario, parse_args, run_scenarios
from grouch.util import issubclass, isinstance

tested_modules = ["grouch.util"]

class SuperEC (Base):
    pass

class ChildEC (SuperEC):
    pass

class Super:
    pass

class Child (Super):
    pass

class ECTest (TestScenario):

    def setup (self):
        self.super_ec = SuperEC()
        self.child_ec = ChildEC()
        self.super = Super()
        self.child = Child()

    def shutdown (self):
        del self.super_ec, self.child_ec
        del self.super, self.child


    test_cases = ['subclass_ec', 'subclass_normal',
                  'instance_ec', 'instance_normal']

    def check_subclass_ec (self):
        "EC classes and subclasses: 7"

        self.test_bool("issubclass(SuperEC, SuperEC)")
        self.test_bool("issubclass(SuperEC, Base)")

        self.test_bool("issubclass(ChildEC, ChildEC)")
        self.test_bool("issubclass(ChildEC, SuperEC)")
        self.test_bool("issubclass(ChildEC, Base)")

        self.test_bool("issubclass(SuperEC, ChildEC)", want_true=0)
        self.test_bool("issubclass(Base, ChildEC)", want_true=0)

    def check_subclass_normal (self):
        "normal classes and subclasses: 11"

        self.test_bool("issubclass(Super, Super)")
        self.test_bool("issubclass(Super, Base)", want_true=0)

        self.test_bool("issubclass(Child, Child)")
        self.test_bool("issubclass(Child, Super)")
        self.test_bool("issubclass(Child, Base)", want_true=0)

        self.test_bool("issubclass(Super, Child)", want_true=0)
        self.test_bool("issubclass(Base, Child)", want_true=0)
        
        self.test_bool("issubclass(Super, SuperEC)", want_true=0)
        self.test_bool("issubclass(Child, SuperEC)", want_true=0)

        self.test_bool("issubclass(ChildEC, Super)", want_true=0)
        self.test_bool("issubclass(ChildEC, Child)", want_true=0)
        
    def check_instance_ec (self):
        "EC instances: 9"

        self.test_bool("isinstance(self.super_ec, SuperEC)")
        self.test_bool("isinstance(self.super_ec, Base)")
        self.test_bool("isinstance(self.super_ec, Super)", want_true=0)
        self.test_bool("isinstance(self.super_ec, ChildEC)", want_true=0)

        self.test_bool("isinstance(self.child_ec, ChildEC)")
        self.test_bool("isinstance(self.child_ec, SuperEC)")
        self.test_bool("isinstance(self.child_ec, Base)")
        self.test_bool("isinstance(self.child_ec, Super)", want_true=0)
        self.test_bool("isinstance(self.child_ec, Child)", want_true=0)


    def check_instance_normal (self):
        "normal instances: 9"

        self.test_bool("isinstance(self.super, Super)")
        self.test_bool("isinstance(self.super, Base)", want_true=0)
        self.test_bool("isinstance(self.super, SuperEC)", want_true=0)
        self.test_bool("isinstance(self.super, ChildEC)", want_true=0)

        self.test_bool("isinstance(self.child, Child)")
        self.test_bool("isinstance(self.child, Super)")
        self.test_bool("isinstance(self.child, Base)", want_true=0)
        self.test_bool("isinstance(self.child, SuperEC)", want_true=0)
        self.test_bool("isinstance(self.child, ChildEC)", want_true=0)


# class ECTest


if __name__ == "__main__":
    (scenarios, options) = parse_args()
    run_scenarios (scenarios, options)



#!/usr/bin/env python

__revision__ = "$Id: setup.py,v 1.3 2001/09/05 14:40:11 gward Exp $"

from distutils.core import setup

setup(name = "Grouch",
      version = "0.1",
      description = "System for describing and enforcing "
                    "a Python object schema",
      long_description = """\
An object schema describes every class in a collection of objects -- in
particular, it specifies the type of every instance attribute of every
class.  Grouch includes a type language for specifying attribute types,
a tool to parse specially-formatted class docstrings and output a
complete object schema, and another tool for walking a persistent object
graph and ensuring that every scrap of data in it conforms to the object
schema extracted from your class docstrings.  The API for defining,
querying, and enforcing types is fairly complete and well-documented, so
you can use Grouch's type system in other ways as well.
""",
      author = "Greg Ward",
      author_email = "gward@mems-exchange.org",
      url = "http://www.mems-exchange.org/software/grouch/",

      package_dir = {"grouch": "lib"},
      packages = ["grouch"],
      scripts = ["scripts/gen_schema", "scripts/check_data"],
     )
